import string

from pyramid_mailer import get_mailer
from pyramid_mailer.message import Message


DEFAULT_EMAIL_SENDER = "admin@example.com"
DEFAULT_SUBJECT_TEMPLATE = "activate your account"
DEFAULT_BODY_TEMPLATE = "{activation-key}"
DEFAULT_CONFIRMATION_SUBJECT_TEMPLATE = "Account active"
DEFAULT_CONFIRMATION_BODY_TEMPLATE = "The account {id} is now active"
DEFAULT_RESET_SUBJECT_TEMPLATE = "Reset password"
DEFAULT_RESET_BODY_TEMPLATE = "{reset-password}"


class EmailFormatter(string.Formatter):
    """Formatter class that will not fail if there's a missing key."""

    def __init__(self, default="{{{0}}}"):
        self.default = default

    def get_value(self, key, args, kwargs):
        return kwargs.get(key, self.default.format(key))


class Emailer:
    def __init__(self, request, user):
        self.request = request
        self.settings = request.registry.settings
        self.user = user
        self.user_email = user["id"]
        self.email_sender = self.settings.get(
            "account_validation.email_sender", DEFAULT_EMAIL_SENDER
        )
        self.mailer = get_mailer(request)

    def send_mail(self, subject_template, body_template, extra_data=None):
        formatter = EmailFormatter()
        if extra_data is None:
            extra_data = {}
        user_email_context = self.user.get("email-context", {})
        # We might have some previous email context.
        try:
            data = self.request.json.get("data", {})
            email_context = data.get("email-context", user_email_context)
        except ValueError:
            email_context = user_email_context

        formatted_subject = formatter.format(
            subject_template, **self.user, **extra_data, **email_context
        )
        formatted_body = formatter.format(
            body_template, **self.user, **extra_data, **email_context
        )
        message = Message(
            subject=formatted_subject,
            sender=self.email_sender,
            recipients=[self.user_email],
            body=formatted_body,
        )
        self.mailer.send(message)

    def send_activation(self, activation_key):
        extra_data = {"activation-key": activation_key}

        subject_template = self.settings.get(
            "account_validation.email_subject_template", DEFAULT_SUBJECT_TEMPLATE
        )
        body_template = self.settings.get(
            "account_validation.email_body_template", DEFAULT_BODY_TEMPLATE
        )
        self.send_mail(subject_template, body_template, extra_data)

    def send_confirmation(self):
        subject_template = self.settings.get(
            "account_validation.email_confirmation_subject_template",
            DEFAULT_CONFIRMATION_SUBJECT_TEMPLATE,
        )
        body_template = self.settings.get(
            "account_validation.email_confirmation_body_template",
            DEFAULT_CONFIRMATION_BODY_TEMPLATE,
        )

        self.send_mail(subject_template, body_template)

    def send_temporary_reset_password(self, reset_password):
        extra_data = {"reset-password": reset_password}
        subject_template = self.settings.get(
            "account_validation.email_reset_password_subject_template",
            DEFAULT_RESET_SUBJECT_TEMPLATE,
        )
        body_template = self.settings.get(
            "account_validation.email_reset_password_body_template", DEFAULT_RESET_BODY_TEMPLATE
        )

        self.send_mail(subject_template, body_template, extra_data)



import logging
import random
import re
import warnings
from datetime import datetime
from secrets import token_hex

from dateutil import parser as dateparser
from pyramid.events import ApplicationCreated, NewRequest, NewResponse
from pyramid.exceptions import ConfigurationError
from pyramid.httpexceptions import (
    HTTPBadRequest,
    HTTPGone,
    HTTPMethodNotAllowed,
    HTTPTemporaryRedirect,
)
from pyramid.interfaces import IAuthenticationPolicy
from pyramid.renderers import JSON as JSONRenderer
from pyramid.response import Response
from pyramid.security import NO_PERMISSION_REQUIRED
from pyramid.settings import asbool, aslist
from pyramid_multiauth import MultiAuthenticationPolicy, MultiAuthPolicySelected

from kinto.core import cache, errors, permission, storage, utils
from kinto.core.events import ACTIONS, ResourceChanged, ResourceRead


try:
    import newrelic.agent
except ImportError:  # pragma: no cover
    newrelic = None
try:
    from werkzeug.middleware.profiler import ProfilerMiddleware
except ImportError:  # pragma: no cover
    ProfilerMiddleware = False
try:
    import sentry_sdk
    from sentry_sdk.integrations.logging import LoggingIntegration
    from sentry_sdk.integrations.pyramid import PyramidIntegration
    from sentry_sdk.integrations.sqlalchemy import SqlalchemyIntegration
except ImportError:  # pragma: no cover
    sentry_sdk = None


logger = logging.getLogger(__name__)
summary_logger = logging.getLogger("request.summary")


def setup_request_bound_data(config):
    """Attach custom data on request object, and share it with parent
    requests during batch."""

    def attach_bound_data(request):
        parent = getattr(request, "parent", None)
        return parent.bound_data if parent else {}

    config.add_request_method(attach_bound_data, name="bound_data", reify=True)


def setup_json_serializer(config):
    import requests
    import webob

    # Monkey patch to use rapidjson
    webob.request.json = utils.json
    requests.models.json = utils.json

    # Override json renderer using rapidjson
    renderer = JSONRenderer(serializer=utils.json_serializer)
    config.add_renderer("ultrajson", renderer)  # See `kinto.core.Service`


def setup_csp_headers(config):
    """Content-Security-Policy HTTP response header helps reduce XSS risks on
    modern browsers by declaring, which dynamic resources are allowed to load.
    On APIs, we disable everything.
    """
    disable_all = "default-src 'none'; frame-ancestors 'none'; base-uri 'none'; "

    def on_new_response(event):
        event.response.headers.setdefault("Content-Security-Policy", disable_all)

    config.add_subscriber(on_new_response, NewResponse)


def restrict_http_methods_if_readonly(config):
    """Prevent write operations if server is configured as read-only.

    This is an additional layer of security on top of the verifications done
    in :method:`kinto.core.Resource.register_service`, in case an installed
    plugin does not take this setting into account in the definition of its
    ad-hoc Pyramid views.
    """
    settings = config.get_settings()

    def on_new_request(event):
        if event.request.method.lower() not in ("get", "head", "options"):
            raise HTTPMethodNotAllowed()

    if asbool(settings["readonly"]):
        config.add_subscriber(on_new_request, NewRequest)


def setup_version_redirection(config):
    """Add a view which redirects to the current version of the API."""
    settings = config.get_settings()
    redirect_enabled = settings["version_prefix_redirect_enabled"]
    version_prefix_redirection_enabled = asbool(redirect_enabled)
    cache_seconds = int(settings["version_prefix_redirect_ttl_seconds"])

    route_prefix = config.route_prefix
    config.registry.route_prefix = route_prefix

    # Redirect to the current version of the API if the prefix isn't used.
    # Do not redirect if kinto.version_prefix_redirect_enabled is set to
    # False.
    if not version_prefix_redirection_enabled:
        return

    def _redirect_to_version_view(request):
        if request.method.lower() == "options":
            # CORS responses should always have status 200.
            return utils.reapply_cors(request, Response())

        querystring = request.url[(request.url.rindex(request.path) + len(request.path)) :]
        redirect = f"/{route_prefix}{request.path}{querystring}"
        resp = HTTPTemporaryRedirect(redirect)
        if cache_seconds >= 0:
            resp.cache_expires(cache_seconds)
        raise resp

    # Disable the route prefix passed by the app.
    config.route_prefix = None

    config.add_route(name="redirect_to_version", pattern=r"/{path:(?!v[0-9]+)[^\r\n]*}")

    config.add_view(
        view=_redirect_to_version_view,
        route_name="redirect_to_version",
        permission=NO_PERMISSION_REQUIRED,
    )

    config.route_prefix = route_prefix


def setup_authentication(config):
    """Let pyramid_multiauth manage authentication and authorization
    from configuration.
    """
    config.include("pyramid_multiauth")
    settings = config.get_settings()

    policies = aslist(settings["multiauth.policies"])
    if "basicauth" in policies:
        config.include("kinto.core.authentication")

    # Track policy used, for prefixing user_id and for logging.
    def on_policy_selected(event):
        request = event.request
        # If the authn policy has a name attribute, we use it rather
        # than using the one specified in settings.
        authn_type = getattr(event.policy, "name", event.policy_name.lower())
        request.authn_type = authn_type
        request.selected_userid = event.userid
        # Add authentication info to context.
        request.log_context(uid=event.userid, authn_type=authn_type)

    config.add_subscriber(on_policy_selected, MultiAuthPolicySelected)


def setup_backoff(config):
    """Attach HTTP requests/responses objects.

    This is useful to attach objects to the request object for easier
    access, and to pre-process responses.
    """

    def on_new_response(event):
        # Add backoff in response headers.
        backoff = config.registry.settings["backoff"]
        if backoff is not None:
            backoff_percentage = config.registry.settings["backoff_percentage"]
            if backoff_percentage is not None:
                if random.random() < (float(backoff_percentage) / 100.0):
                    event.response.headers["Backoff"] = str(backoff)
            else:
                event.response.headers["Backoff"] = str(backoff)

    config.add_subscriber(on_new_response, NewResponse)


def setup_requests_scheme(config):
    """Force server scheme, host and port at the application level."""
    settings = config.get_settings()

    http_scheme = settings["http_scheme"]
    http_host = settings["http_host"]

    def on_new_request(event):
        if http_scheme:
            event.request.scheme = http_scheme
        if http_host:
            event.request.host = http_host

    if http_scheme or http_host:
        config.add_subscriber(on_new_request, NewRequest)


def setup_deprecation(config):
    config.add_tween("kinto.core.initialization._end_of_life_tween_factory")


def _end_of_life_tween_factory(handler, registry):
    """Pyramid tween to handle service end of life."""
    deprecation_msg = "The service you are trying to connect no longer exists" " at this location."

    def eos_tween(request):
        eos_date = registry.settings["eos"]
        eos_url = registry.settings["eos_url"]
        eos_message = registry.settings["eos_message"]
        if not eos_date:
            return handler(request)

        eos_date = dateparser.parse(eos_date)
        if eos_date > datetime.now():
            code = "soft-eol"
            request.response = handler(request)
        else:
            code = "hard-eol"
            request.response = errors.http_error(
                HTTPGone(), errno=errors.ERRORS.SERVICE_DEPRECATED, message=deprecation_msg
            )

        errors.send_alert(request, eos_message, url=eos_url, code=code)
        return utils.reapply_cors(request, request.response)

    return eos_tween


def setup_storage(config):
    settings = config.get_settings()

    # Id generators by resource name.
    config.registry.id_generators = {}
    for key, value in settings.items():
        m = re.match(r"^([^_]*)_?id_generator", key)
        if m is None:
            continue
        resource_name = m.group(1)
        id_generator = config.maybe_dotted(value)
        config.registry.id_generators[resource_name] = id_generator()

    storage_mod = settings["storage_backend"]
    if not storage_mod:
        return

    storage_mod = config.maybe_dotted(storage_mod)
    backend = storage_mod.load_from_config(config)
    if not isinstance(backend, storage.StorageBase):
        raise ConfigurationError(f"Invalid storage backend: {backend}")
    config.registry.storage = backend

    heartbeat = storage.heartbeat(backend)
    config.registry.heartbeats["storage"] = heartbeat


def setup_permission(config):
    settings = config.get_settings()
    permission_mod = settings["permission_backend"]
    if not permission_mod:
        return

    permission_mod = config.maybe_dotted(permission_mod)
    backend = permission_mod.load_from_config(config)
    if not isinstance(backend, permission.PermissionBase):
        raise ConfigurationError(f"Invalid permission backend: {backend}")
    config.registry.permission = backend

    heartbeat = permission.heartbeat(backend)
    config.registry.heartbeats["permission"] = heartbeat


def setup_cache(config):
    settings = config.get_settings()
    cache_mod = settings["cache_backend"]
    if not cache_mod:
        return

    cache_mod = config.maybe_dotted(cache_mod)
    backend = cache_mod.load_from_config(config)
    if not isinstance(backend, cache.CacheBase):
        raise ConfigurationError(f"Invalid cache backend: {backend}")
    config.registry.cache = backend

    heartbeat = cache.heartbeat(backend)
    config.registry.heartbeats["cache"] = heartbeat


def setup_sentry(config):
    settings = config.get_settings()

    # Note: SENTRY_DSN and SENTRY_ENV env variables will override
    # .ini values thanks to `load_default_settings()`.

    dsn = settings["sentry_dsn"]
    if dsn:  # pragma: no cover
        env_options = {}
        env = settings["sentry_env"]
        if env:
            env_options["environment"] = env

        breadcrumbs_level = settings["sentry_breadcrumbs_min_level"]
        events_level = settings["sentry_events_min_level"]
        sentry_sdk.init(
            dsn,
            integrations=[
                PyramidIntegration(),
                SqlalchemyIntegration(),
                LoggingIntegration(
                    # Logs to be captured as breadcrumbs (debug and above by default)
                    level=breadcrumbs_level,
                    # Logs to be catpured as events (warning and above by default)
                    event_level=events_level,
                ),
            ],
            **env_options,
        )

        def on_app_created(event):
            msg = "Running {project_name} {project_version}.".format_map(settings)
            sentry_sdk.capture_message(msg, "info")

        config.add_subscriber(on_app_created, ApplicationCreated)


def setup_statsd(config):
    settings = config.get_settings()
    config.registry.statsd = None

    if settings["statsd_url"]:
        statsd_mod = settings["statsd_backend"]
        statsd_mod = config.maybe_dotted(statsd_mod)
        client = statsd_mod.load_from_config(config)

        config.registry.statsd = client

        client.watch_execution_time(config.registry.cache, prefix="backend")
        client.watch_execution_time(config.registry.storage, prefix="backend")
        client.watch_execution_time(config.registry.permission, prefix="backend")

        # Commit so that configured policy can be queried.
        config.commit()
        policy = config.registry.queryUtility(IAuthenticationPolicy)
        if isinstance(policy, MultiAuthenticationPolicy):
            for name, subpolicy in policy.get_policies():
                client.watch_execution_time(subpolicy, prefix="authentication", classname=name)
        else:
            client.watch_execution_time(policy, prefix="authentication")

        def on_new_response(event):
            request = event.request

            # Count unique users.
            user_id = request.prefixed_userid
            if user_id:
                # Get rid of colons in metric packet (see #1282).
                user_id = user_id.replace(":", ".")
                client.count("users", unique=user_id)

            # Count authentication verifications.
            if hasattr(request, "authn_type"):
                client.count(f"authn_type.{request.authn_type}")

            # Count view calls.
            service = request.current_service
            if service:
                client.count(f"view.{service.name}.{request.method}")

        config.add_subscriber(on_new_response, NewResponse)

        return client


def install_middlewares(app, settings):
    "Install a set of middlewares defined in the ini file on the given app."
    # Setup new-relic.
    if settings.get("newrelic_config"):
        ini_file = settings["newrelic_config"]
        env = settings["newrelic_env"]
        newrelic.agent.initialize(ini_file, env)
        app = newrelic.agent.WSGIApplicationWrapper(app)

    # Adds the Werkzeug profiler.
    if asbool(settings.get("profiler_enabled")):
        profile_dir = settings["profiler_dir"]
        app = ProfilerMiddleware(app, profile_dir=profile_dir, restrictions=("*kinto.core*"))

    return app


def setup_logging(config):
    """Setup structured logging, and emit `request.summary` event on each
    request, as recommended by Mozilla Services standard:

    * https://mana.mozilla.org/wiki/display/CLOUDSERVICES/Logging+Standard
    * http://12factor.net/logs
    """

    def on_new_request(event):
        request = event.request
        # Save the time the request was received by the server.
        event.request._received_at = utils.msec_time()

        try:
            # Pyramid fails if the URL contains invalid UTF-8 characters.
            request_path = event.request.path
        except UnicodeDecodeError:
            raise errors.http_error(
                HTTPBadRequest(),
                errno=errors.ERRORS.INVALID_PARAMETERS,
                message="Invalid URL path.",
            )

        request.log_context(
            agent=request.headers.get("User-Agent"),
            path=request_path,
            method=request.method,
            lang=request.headers.get("Accept-Language"),
            rid=request.headers.get("X-Request-Id", token_hex(16)),
            errno=0,
        )
        qs = dict(errors.request_GET(request))
        if qs:
            request.log_context(querystring=qs)

        if summary_logger.level == logging.DEBUG:
            request.log_context(headers=dict(request.headers), body=request.body)

    config.add_subscriber(on_new_request, NewRequest)

    def on_new_response(event):
        response = event.response
        request = event.request

        # Compute the request processing time in msec (-1 if unknown)
        current = utils.msec_time()
        duration = current - getattr(request, "_received_at", current - 1)
        isotimestamp = datetime.fromtimestamp(current / 1000).isoformat()

        # Bind infos for request summary logger.
        request.log_context(time=isotimestamp, code=response.status_code, t=duration)

        if summary_logger.level == logging.DEBUG:
            request.log_context(response=dict(headers=dict(response.headers), body=response.body))

        try:
            # If error response, bind errno.
            request.log_context(errno=response.errno)
        except AttributeError:
            pass

        if not hasattr(request, "parent"):
            # Ouput application request summary.
            summary_logger.info("", extra=request.log_context())

    config.add_subscriber(on_new_response, NewResponse)


class EventActionFilter:
    def __init__(self, actions, config):
        actions = ACTIONS.from_string_list(actions)
        self.actions = [action.value for action in actions]

    def phash(self):
        return f"for_actions = {','.join(self.actions)}"

    def __call__(self, event):
        action = event.payload.get("action")
        return not action or action in self.actions


class EventResourceFilter:
    def __init__(self, resources, config):
        self.resources = resources

    def phash(self):
        return f"for_resources = {','.join(self.resources)}"

    def __call__(self, event):
        resource = event.payload.get("resource_name")
        return not resource or not self.resources or resource in self.resources


def setup_listeners(config):
    # Register basic subscriber predicates, to filter events.
    config.add_subscriber_predicate("for_actions", EventActionFilter)
    config.add_subscriber_predicate("for_resources", EventResourceFilter)

    write_actions = (ACTIONS.CREATE, ACTIONS.UPDATE, ACTIONS.DELETE)
    settings = config.get_settings()
    settings_prefix = settings.get("settings_prefix", "")
    listeners = aslist(settings["event_listeners"])

    for name in listeners:
        logger.info(f"Setting up '{name}' listener")
        prefix = f"event_listeners.{name}."

        try:
            listener_mod = config.maybe_dotted(name)
            prefix = f"event_listeners.{name.split('.')[-1]}."
            listener = listener_mod.load_from_config(config, prefix)
        except (ImportError, AttributeError):
            module_setting = prefix + "use"
            # Read from ENV or settings.
            module_value = utils.read_env(
                f"{settings_prefix}.{module_setting}", settings.get(module_setting)
            )
            listener_mod = config.maybe_dotted(module_value)
            listener = listener_mod.load_from_config(config, prefix)

        # If StatsD is enabled, monitor execution time of listeners.
        if getattr(config.registry, "statsd", None):
            statsd_client = config.registry.statsd
            key = f"listeners.{name}"
            listener = statsd_client.timer(key)(listener.__call__)

        # Optional filter by event action.
        actions_setting = prefix + "actions"
        # Read from ENV or settings.
        actions_value = utils.read_env(
            f"{settings_prefix}.{actions_setting}", settings.get(actions_setting, "")
        )
        actions = aslist(actions_value)
        if len(actions) > 0:
            actions = ACTIONS.from_string_list(actions)
        else:
            actions = write_actions

        # Optional filter by event resource name.
        resource_setting = prefix + "resources"
        # Read from ENV or settings.
        resource_value = utils.read_env(
            f"{settings_prefix}.{resource_setting}", settings.get(resource_setting, "")
        )
        resource_names = aslist(resource_value)

        # Pyramid event predicates.
        options = dict(for_actions=actions, for_resources=resource_names)

        if ACTIONS.READ in actions:
            config.add_subscriber(listener, ResourceRead, **options)
            actions = [a for a in actions if a != ACTIONS.READ]

        if len(actions) > 0:
            config.add_subscriber(listener, ResourceChanged, **options)


def load_default_settings(config, default_settings):
    """Read settings provided in Paste ini file, set default values and
    replace if defined as environment variable.
    """
    settings = config.get_settings()

    settings_prefix = settings["settings_prefix"]

    def _prefixed_keys(key):
        unprefixed = key
        if key.startswith(settings_prefix + "."):
            unprefixed = key.split(".", 1)[1]
        project_prefix = f"{settings_prefix}.{unprefixed}"
        return unprefixed, project_prefix

    # Fill settings with default values if not defined.
    for key, default_value in sorted(default_settings.items()):
        unprefixed, project_prefix = keys = _prefixed_keys(key)
        is_defined = len(set(settings.keys()).intersection(set(keys))) > 0
        if not is_defined:
            settings[unprefixed] = default_value

    for key, value in sorted(settings.items()):
        value = utils.native_value(value)
        unprefixed, project_prefix = keys = _prefixed_keys(key)

        # Fail if not only one is defined.
        defined = set(settings.keys()).intersection(set(keys))
        distinct_values = set([str(settings[d]) for d in defined])

        if len(defined) > 1 and len(distinct_values) > 1:
            names = "', '".join(defined)
            raise ValueError(f"Settings '{names}' are in conflict.")

        # Override settings from OS env values.
        # e.g. HTTP_PORT, READINGLIST_HTTP_PORT, KINTO_HTTP_PORT
        from_env = utils.read_env(unprefixed, value)
        from_env = utils.read_env(project_prefix, from_env)

        settings[unprefixed] = from_env

    config.add_settings(settings)


def initialize(config, version=None, settings_prefix="", default_settings=None):
    """Initialize kinto.core with the given configuration, version and project
    name.

    This will basically include kinto.core in Pyramid and set route prefix
    based on the specified version.

    :param config: Pyramid configuration
    :type config: :class:`~pyramid:pyramid.config.Configurator`
    :param str version: Current project version (e.g. '0.0.1') if not defined
        in application settings.
    :param str settings_prefix: Project name if not defined
        in application settings.
    :param dict default_settings: Override kinto.core default settings values.
    """
    from kinto.core import DEFAULT_SETTINGS

    settings = config.get_settings()

    settings_prefix = (
        settings.pop("kinto.settings_prefix", settings.get("settings_prefix")) or settings_prefix
    )
    settings["settings_prefix"] = settings_prefix
    if not settings_prefix:
        warnings.warn("No value specified for `settings_prefix`")

    kinto_core_defaults = {**DEFAULT_SETTINGS}

    if default_settings:
        kinto_core_defaults.update(default_settings)

    load_default_settings(config, kinto_core_defaults)

    http_scheme = settings["http_scheme"]
    if http_scheme != "https":
        warnings.warn("HTTPS is not enabled")

    # Override project version from settings.
    project_version = settings.get("project_version") or version
    if not project_version:
        error_msg = f"Invalid project version: {project_version}"
        raise ConfigurationError(error_msg)
    settings["project_version"] = project_version = str(project_version)

    # HTTP API version.
    http_api_version = settings.get("http_api_version")
    if http_api_version is None:
        # The API version is derivated from the module version if not provided.
        http_api_version = ".".join(project_version.split(".")[0:2])
    settings["http_api_version"] = http_api_version = str(http_api_version)
    api_version = f"v{http_api_version.split('.')[0]}"

    # Include kinto.core views with the correct api version prefix.
    config.include("kinto.core", route_prefix=api_version)
    config.route_prefix = api_version



from unittest import mock

from pyramid import testing

from kinto.core.testing import unittest

from .support import BaseWebTest


class HelloViewTest(BaseWebTest, unittest.TestCase):
    def test_returns_info_about_url_and_version(self):
        response = self.app.get("/")
        self.assertEqual(response.json["project_version"], "0.0.1")
        self.assertEqual(response.json["url"], "http://localhost/v0/")
        self.assertEqual(response.json["project_name"], "myapp")
        self.assertEqual(response.json["project_docs"], "https://kinto.readthedocs.io/")

    def test_does_not_escape_forward_slashes(self):
        response = self.app.get("/")
        self.assertNotIn("\\/", str(response.body))

    def test_do_not_returns_eos_if_empty_in_settings(self):
        response = self.app.get("/")
        self.assertNotIn("eos", response.json)

    def test_returns_eos_if_not_empty_in_settings(self):
        eos = "2069-02-21"
        with mock.patch.dict(self.app.app.registry.settings, [("eos", eos)]):
            response = self.app.get("/")
            self.assertEqual(response.json["eos"], eos)

    def test_public_settings_are_shown_in_view(self):
        response = self.app.get("/")
        settings = response.json["settings"]
        expected = {"batch_max_requests": 25, "readonly": False, "explicit_permissions": True}
        self.assertEqual(expected, settings)

    def test_public_settings_can_be_set_from_registry(self):
        self.app.app.registry.public_settings.add("paginate_by")
        response = self.app.get("/")
        settings = response.json["settings"]
        self.assertIn("paginate_by", settings)

    def test_if_user_not_authenticated_no_userid_provided(self):
        response = self.app.get("/")
        self.assertNotIn("user", response.json)

    def test_if_user_authenticated_userid_is_provided(self):
        response = self.app.get("/", headers=self.headers)
        userid = response.json["user"]["id"]
        self.assertTrue(
            userid.startswith("basicauth:"), '"{}" does not start with "basicauth:"'.format(userid)
        )

    def test_if_user_authenticated_profile_is_provided_if_present(self):
        settings = self.get_app_settings()
        config = testing.setUp(settings=settings)
        app = self.make_app(config=config)
        profile = {
            "email": "foobar@domain.tld",
            "profile_pic": "https://flatimage.com/100x100.png",
        }

        def get_profile(request):
            return profile

        config.add_request_method(get_profile, name="get_user_profile")

        response = app.get("/", headers=self.headers)
        self.assertEqual(response.json["user"]["profile"], profile)

    def test_return_http_api_version_when_set(self):
        with mock.patch.dict(self.app.app.registry.settings, [("http_api_version", "1.2")]):
            response = self.app.get("/")

        self.assertTrue(response.json["http_api_version"])


class APICapabilitiesTest(BaseWebTest, unittest.TestCase):
    def test_list_of_capabilities_contains_basicauth_by_default(self):
        app = self.make_app()
        response = app.get("/")
        capabilities = response.json["capabilities"]
        self.assertEqual(
            capabilities,
            {
                "basicauth": {
                    "description": "Very basic authentication sessions. Not for production use.",
                    "url": "http://kinto.readthedocs.io/en/stable/api/1.x/authentication.html",
                }
            },
        )

    def test_capabilities_can_be_specified_via_config(self):
        settings = self.get_app_settings()
        settings["multiauth.policies"] = ""
        config = testing.setUp(settings=settings)
        app = self.make_app(config=config)

        config.add_api_capability("cook-coffee")

        response = app.get("/")
        capabilities = response.json["capabilities"]
        expected = {"cook-coffee": {"description": "", "url": ""}}
        self.assertEqual(capabilities, expected)

    def test_capabilities_can_have_arbitrary_attributes(self):
        config = testing.setUp(settings=self.get_app_settings())
        app = self.make_app(config=config)

        capability = dict(
            description="Track object change",
            url="https://github.com/Kinto/kinto-changes",
            status="beta",
        )
        config.add_api_capability("object-changes", **capability)

        response = app.get("/")
        capabilities = response.json["capabilities"]
        self.assertEqual(capabilities["object-changes"], capability)

    def test_capability_fails_if_already_registered(self):
        config = testing.setUp(settings=self.get_app_settings())
        self.make_app(config=config)

        config.add_api_capability("cook-coffee")
        with self.assertRaises(ValueError):
            config.add_api_capability("cook-coffee")



import functools
import logging

from pyramid.authorization import Authenticated
from pyramid.interfaces import IAuthorizationPolicy
from pyramid.settings import aslist
from zope.interface import implementer

from kinto.core import utils
from kinto.core.storage import exceptions as storage_exceptions


logger = logging.getLogger(__name__)

# When permission is set to "private", only the current user is allowed.
PRIVATE = "private"

# A permission is called "dynamic" when it's computed at request time.
DYNAMIC = "dynamic"


def groupfinder(userid, request):
    """Fetch principals from permission backend for the specified `userid`.

    This is plugged by default using the ``multiauth.groupfinder`` setting.
    """
    backend = getattr(request.registry, "permission", None)
    # Permission backend not configured. Ignore.
    if not backend:
        return []

    # Safety check when Kinto-Core is used without pyramid_multiauth.
    if request.prefixed_userid:
        userid = request.prefixed_userid

    # Query the permission backend only once per request (e.g. batch).
    reify_key = userid + "_principals"
    if reify_key not in request.bound_data:
        principals = backend.get_user_principals(userid)
        request.bound_data[reify_key] = principals

    return request.bound_data[reify_key]


@implementer(IAuthorizationPolicy)
class AuthorizationPolicy:
    """Default authorization class, that leverages the permission backend
    for shareable resources.
    """

    get_bound_permissions = None
    """Callable that takes an object id and a permission and returns
    a list of tuples (<object id>, <permission>). Useful when objects
    permission depend on others."""

    def permits(self, context, principals, permission):
        if permission == PRIVATE:
            # When using the private permission, we bypass the permissions
            # backend, and simply authorize if authenticated.
            return Authenticated in principals

        principals = context.get_prefixed_principals()

        create_permission = f"{context.resource_name}:create"

        permission = context.required_permission
        if permission == "create":
            permission = create_permission

        object_id = context.permission_object_id
        bound_perms = self._get_bound_permissions(object_id, permission)

        allowed = context.check_permission(principals, bound_perms)

        # Here we consider that parent URI is one path level above.
        parent_uri = "/".join(object_id.split("/")[:-1]) if object_id else None

        # If not allowed to delete/patch, and target object is missing, and
        # allowed to read the parent, then view is permitted (will raise 404
        # later anyway). See Kinto/kinto#918
        is_object_unknown = not context.on_plural_endpoint and context.current_object is None
        if context.required_permission == "write" and is_object_unknown:
            bound_perms = self._get_bound_permissions(parent_uri, "read")
            allowed = context.check_permission(principals, bound_perms)

        # If not allowed on this plural endpoint, but some objects are shared with
        # the current user, then authorize.
        # The :class:`kinto.core.resource.Resource` class will take care of the filtering.
        is_list_operation = (
            context.on_plural_endpoint
            and not permission.endswith("create")
            and context.current_object is None
        )
        if not allowed and is_list_operation:
            allowed = bool(
                context.fetch_shared_objects(permission, principals, self.get_bound_permissions)
            )
            if not allowed:
                # If allowed to create this kind of object on parent,
                # then allow to obtain the list.
                if len(bound_perms) > 0:
                    bound_perms = [(parent_uri, create_permission)]
                else:
                    bound_perms = [("", "create")]  # Root object.
                allowed = context.check_permission(principals, bound_perms)

        if not allowed:
            logger.info(
                "Permission %r on %r not granted to %r.",
                permission,
                object_id,
                principals[0],
                extra=dict(userid=principals[0], uri=object_id, perm=permission),
            )

        return allowed

    def _get_bound_permissions(self, object_id, permission):
        if self.get_bound_permissions is None:
            # Permission to 'write' gives permission to 'read'.
            bound = [(object_id, permission)]
            if permission == "read":
                bound += [(object_id, "write")]
            return bound
        return self.get_bound_permissions(object_id, permission)

    def principals_allowed_by_permission(self, context, permission):
        raise NotImplementedError()  # PRAGMA NOCOVER


class RouteFactory:
    resource_name = None
    on_plural_endpoint = False
    required_permission = None
    permission_object_id = None
    current_object = None
    shared_ids = None

    method_permissions = {
        "head": "read",
        "get": "read",
        "post": "create",
        "delete": "write",
        "patch": "write",
    }

    def __init__(self, request):
        # Store some shortcuts.
        permission = request.registry.permission
        self._check_permission = permission.check_permission
        self._get_accessible_objects = permission.get_accessible_objects

        self.get_prefixed_principals = functools.partial(utils.prefixed_principals, request)

        # Store current resource and required permission.
        service = utils.current_service(request)
        is_on_resource = (
            service is not None and hasattr(service, "viewset") and hasattr(service, "resource")
        )
        self._resource = None
        if is_on_resource:
            self.resource_name = request.current_resource_name
            self.on_plural_endpoint = getattr(service, "type", None) == "plural"

            # Check if this request targets an individual object.
            # Its existence will affect permissions checking (cf `_find_required_permission()`).
            # There are cases where the permission is not directly related to the HTTP method,
            # For example:
            # - with POST on plural endpoint, with an id supplied
            # - with PUT on an object, which can either be creation or update
            is_write_on_object = not self.on_plural_endpoint and request.method.lower() in (
                "put",
                "delete",
                "patch",
            )
            is_post_on_plural = self.on_plural_endpoint and request.method.lower() == "post"
            if is_write_on_object or is_post_on_plural:
                # We instantiate the resource to determine the object targeted by the request.
                self._resource = resource = service.resource(request=request, context=self)
                if resource.object_id is not None:  # Skip POST on plural without id.
                    try:
                        # Save a reference, to avoid refetching from storage in resource.
                        self.current_object = resource.model.get_object(resource.object_id)
                    except storage_exceptions.ObjectNotFoundError:
                        pass

            self.permission_object_id, self.required_permission = self._find_required_permission(
                request, service
            )

            # To obtain shared objects on a plural endpoint, use a match:
            self._object_id_match = self.get_permission_object_id(request, "*")

        self._settings = request.registry.settings

    def check_permission(self, principals, bound_perms):
        """Read allowed principals from settings, if not any, query the permission
        backend to check if view is allowed.
        """
        if not bound_perms:
            bound_perms = [(self.resource_name, self.required_permission)]
        for _, permission in bound_perms:
            # With Kinto inheritance tree, we can have: `permission = "record:create"`
            if self.resource_name and permission.startswith(self.resource_name):
                setting = f"{permission.replace(':', '_')}_principals"
            else:
                setting = f"{self.resource_name}_{permission}_principals"
            allowed_principals = aslist(self._settings.get(setting, ""))
            if allowed_principals:
                if bool(set(allowed_principals) & set(principals)):
                    return True
        return self._check_permission(principals, bound_perms)

    def fetch_shared_objects(self, perm, principals, get_bound_permissions):
        """Fetch objects that are readable or writable for the current
        principals.

        See :meth:`kinto.core.authorization.AuthorizationPolicy.permits`

        If no object is shared, it returns None.

        .. warning::
            This sets the ``shared_ids`` attribute to the context with the
            return value. The attribute is then read by
            :class:`kinto.core.resource.Resource`
        """
        if get_bound_permissions:
            bound_perms = get_bound_permissions(self._object_id_match, perm)
        else:
            bound_perms = [(self._object_id_match, perm)]
        by_obj_id = self._get_accessible_objects(principals, bound_perms, with_children=False)
        ids = by_obj_id.keys()
        # Store for later use in ``Resource``.
        self.shared_ids = [self._extract_object_id(id_) for id_ in ids]
        return self.shared_ids

    def get_permission_object_id(self, request, object_id=None):
        """Returns the permission object id for the current request.
        In the nominal case, it is just the current URI without version prefix.
        For plural endpoint, it is the related object URI using the specified
        `object_id`.

        See :meth:`kinto.core.resource.model.SharableModel` and
        :meth:`kinto.core.authorization.RouteFactory.__init__`
        """
        object_uri = utils.strip_uri_prefix(request.path)

        if self.on_plural_endpoint and object_id is not None:
            # With the current request on a plural endpoint, the object URI must
            # be found out by inspecting the "plural" service and its sibling
            # "object" service. (see `register_resource()`)
            matchdict = {**request.matchdict, "id": object_id}
            try:
                object_uri = utils.instance_uri(request, self.resource_name, **matchdict)
                object_uri = object_uri.replace("%2A", "*")
            except KeyError:
                # Maybe the resource has no single object endpoint.
                # We consider that object URIs in permissions backend will
                # be stored naively:
                object_uri = f"{object_uri}/{object_id}"

        return object_uri

    def _extract_object_id(self, object_uri):
        # XXX: Rewrite using kinto.core.utils.view_lookup() and matchdict['id']
        return object_uri.split("/")[-1]

    def _find_required_permission(self, request, service):
        """Find out what is the permission object id and the required
        permission.

        .. note::
            This method saves an attribute ``self.current_object`` used
            in :class:`kinto.core.resource.Resource`.
        """
        # By default, it's a URI a and permission associated to the method.
        permission_object_id = self.get_permission_object_id(request)
        method = request.method.lower()
        required_permission = self.method_permissions.get(method)

        # For create permission, the object id is the plural endpoint.
        plural_path = str(service.plural_path)
        plural_path = plural_path.format_map(request.matchdict)

        # In the case of a "PUT", check if the targetted object already
        # exists, return "write" if it does, "create" otherwise.
        if request.method.lower() == "put":
            if self.current_object is None:
                # The object does not exist, the permission to create on
                # the related plural endpoint is required.
                permission_object_id = plural_path
                required_permission = "create"
            else:
                # For safe creations, the user needs a create permission.
                # See Kinto/kinto#792
                if request.headers.get("If-None-Match") == "*":
                    permission_object_id = plural_path
                    required_permission = "create"
                else:
                    required_permission = "write"

        # In the case of a "POST" on a plural endpoint, if an "id" was
        # specified, then the object is returned. The required permission
        # is thus "read" on this object.
        if request.method.lower() == "post" and self.current_object is not None:
            permission_object_id = self.get_permission_object_id(
                request, object_id=self._resource.object_id
            )
            required_permission = "read"

        return (permission_object_id, required_permission)



import json
import os
import unittest
from unittest import mock

from pyramid import testing

from kinto.core.cache import postgresql as postgresql_cache
from kinto.core.permission import postgresql as postgresql_permission
from kinto.core.storage import postgresql as postgresql_storage
from kinto.core.storage.postgresql.migrator import MigratorMixin
from kinto.core.testing import skip_if_no_postgresql
from kinto.core.utils import sqlalchemy as sa


@mock.patch.object(MigratorMixin, "_execute_sql_file")
class MigratorTest(unittest.TestCase):
    def setUp(self):
        self.migrator = MigratorMixin()
        here = os.path.dirname(__file__)
        migrations_directory = os.path.join(here, "migrations")
        self.migrator.migrations_directory = migrations_directory
        self.migrator.schema_version = 6

    def test_schema_is_created_if_no_version(self, execute_sql):
        with mock.patch.object(self.migrator, "create_schema") as create_schema:
            with mock.patch.object(self.migrator, "get_installed_version") as installed_version:
                installed_version.return_value = None
                self.migrator.create_or_migrate_schema()
        self.assertTrue(create_schema.called)

    def test_schema_is_not_touched_if_already_current(self, execute_sql):
        # Patch to keep track of SQL files executed.
        with mock.patch.object(self.migrator, "get_installed_version") as installed_version:
            installed_version.return_value = 6
            self.migrator.create_or_migrate_schema()
        self.assertFalse(execute_sql.called)

    def _walk_from_3_to_6(self):
        """Helper method for simulating a migration from version 3 to version 6."""
        versions = [6, 5, 4, 3, 3]
        self.migrator.get_installed_version = lambda: versions.pop()

    def test_migration_file_is_executed_for_every_intermediary_version(self, execute_sql):
        self._walk_from_3_to_6()

        self.migrator.create_or_migrate_schema()
        sql_called = execute_sql.call_args_list[-3][0][0]
        self.assertIn(os.path.join("migrations", "migration_003_004.sql"), sql_called)
        sql_called = execute_sql.call_args_list[-2][0][0]
        self.assertIn(os.path.join("migrations", "migration_004_005.sql"), sql_called)
        sql_called = execute_sql.call_args_list[-1][0][0]
        self.assertIn(os.path.join("migrations", "migration_005_006.sql"), sql_called)

    def test_migration_files_are_listed_if_ran_with_dry_run(self, execute_sql):
        self._walk_from_3_to_6()

        with mock.patch("kinto.core.storage.postgresql.migrator.logger") as mocked:
            self.migrator.create_or_migrate_schema(dry_run=True)

        output = "".join([repr(call) for call in mocked.info.call_args_list]).replace("\\\\", "\\")
        self.assertIn(os.path.join("migrations", "migration_003_004.sql"), output)
        self.assertIn(os.path.join("migrations", "migration_004_005.sql"), output)
        self.assertIn(os.path.join("migrations", "migration_005_006.sql"), output)

    def test_migration_fails_if_intermediary_version_is_missing(self, execute_sql):
        with mock.patch.object(self.migrator, "get_installed_version") as current:
            current.return_value = -1
            self.assertRaises(AssertionError, self.migrator.create_or_migrate_schema)


@skip_if_no_postgresql
class PostgresqlStorageMigrationTest(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        from kinto.core.utils import sqlalchemy

        if sqlalchemy is None:
            return

        from .test_storage import PostgreSQLStorageTest

        self.settings = {**PostgreSQLStorageTest.settings}
        self.config = testing.setUp()
        self.config.add_settings(self.settings)
        self.version = postgresql_storage.Storage.schema_version
        # Usual storage object to manipulate the storage.
        self.storage = postgresql_storage.load_from_config(self.config)

    def setUp(self):
        # Start empty.
        self._delete_everything()

    def tearDown(self):
        postgresql_storage.Storage.schema_version = self.version
        # Finish empty.
        self._delete_everything()

    def _delete_everything(self):
        q = """
        DROP TABLE IF EXISTS records CASCADE;
        DROP TABLE IF EXISTS objects CASCADE;
        DROP TABLE IF EXISTS deleted CASCADE;
        DROP TABLE IF EXISTS timestamps CASCADE;
        DROP TABLE IF EXISTS metadata CASCADE;
        DROP FUNCTION IF EXISTS resource_timestamp(VARCHAR, VARCHAR);
        DROP FUNCTION IF EXISTS collection_timestamp(VARCHAR, VARCHAR);
        DROP FUNCTION IF EXISTS bump_timestamp();
        """
        with self.storage.client.connect() as conn:
            conn.execute(sa.text(q))

    def _load_schema(self, filepath):
        with self.storage.client.connect() as conn:
            here = os.path.dirname(__file__)
            with open(os.path.join(here, filepath)) as f:
                old_schema = f.read()
            conn.execute(sa.text(old_schema))

    def test_does_not_execute_if_ran_with_dry(self):
        self.storage.initialize_schema(dry_run=True)
        query = """SELECT 1 FROM information_schema.tables
        WHERE table_name = 'objects';"""
        with self.storage.client.connect(readonly=True) as conn:
            result = conn.execute(sa.text(query))
        self.assertEqual(result.rowcount, 0)

    def test_schema_sets_the_current_version(self):
        # Create schema in its last version
        self.storage.initialize_schema()
        version = self.storage.get_installed_version()
        self.assertEqual(version, self.version)

    def test_schema_is_considered_first_version_if_no_version_detected(self):
        # Create schema in its last version
        self.storage.initialize_schema()
        with self.storage.client.connect() as conn:
            q = "DELETE FROM metadata WHERE name = 'storage_schema_version';"
            conn.execute(sa.text(q))

        self.assertEqual(self.storage.get_installed_version(), 1)

    def test_schema_is_considered_20_if_server_is_wiped(self):
        # Create schema in its last version
        self.storage.initialize_schema()
        with self.storage.client.connect() as conn:
            q = "DELETE FROM metadata;"
            conn.execute(sa.text(q))

        self.assertEqual(self.storage.get_installed_version(), 20)

    def test_every_available_migration(self):
        """Test every migration available in kinto.core code base since
        version 1.6.

        Objects migration test is currently very naive, and should be
        elaborated along future migrations.
        """

        # Install old schema
        self._load_schema("schema/postgresql-storage-1.6.sql")

        # Create a sample object using some code that is compatible with the
        # schema in place in cliquet 1.6.
        with self.storage.client.connect() as conn:
            before = {"drink": "cacao"}
            query = """
            INSERT INTO records (user_id, resource_name, data)
            VALUES (:user_id, :resource_name, (:data)::JSON)
            RETURNING id, as_epoch(last_modified) AS last_modified;
            """
            placeholders = dict(
                user_id="jean-louis", resource_name="test", data=json.dumps(before)
            )
            result = conn.execute(sa.text(query), placeholders)
            inserted = result.fetchone()
            before["id"] = str(inserted.id)
            before["last_modified"] = inserted.last_modified

        # In cliquet 1.6, version = 1.
        version = self.storage.get_installed_version()
        self.assertEqual(version, 1)

        # Run every migrations available.
        self.storage.initialize_schema()

        # Version matches current one.
        version = self.storage.get_installed_version()
        self.assertEqual(version, self.version)

        # Check that previously created object is still here
        migrated = self.storage.list_all("test", "jean-louis")
        self.assertEqual(migrated[0], before)

        # Check that new objects can be created
        r = self.storage.create("test", ",jean-louis", {"drink": "mate"})

        # And deleted
        self.storage.delete("test", ",jean-louis", r["id"])

    def test_every_available_migration_succeeds_if_tables_were_flushed(self):
        self.storage.initialize_schema()
        # During tests, tables can be flushed.
        self.storage.flush()
        self.storage.initialize_schema()
        # Version matches current one.
        version = self.storage.get_installed_version()
        self.assertEqual(version, self.version)

    def test_migration_12_clean_tombstones(self):
        last_version = postgresql_storage.Storage.schema_version
        postgresql_storage.Storage.schema_version = 11

        self._load_schema("schema/postgresql-storage-11.sql")

        insert_query = """
        INSERT INTO records (id, parent_id, collection_id, data, last_modified)
        VALUES (:id, :parent_id, :collection_id, (:data)::JSONB, from_epoch(:last_modified))
        """
        placeholders = dict(
            id="rid",
            parent_id="jean-louis",
            collection_id="test",
            data=json.dumps({"drink": "mate"}),
            last_modified=123456,
        )
        with self.storage.client.connect() as conn:
            conn.execute(sa.text(insert_query), placeholders)

        create_tombstone = """
        INSERT INTO deleted (id, parent_id, collection_id, last_modified)
        VALUES (:id, :parent_id, :collection_id, from_epoch(:last_modified))
        """
        with self.storage.client.connect() as conn:
            conn.execute(sa.text(create_tombstone), placeholders)

        # Execute the 011 to 012 migration (and others)
        postgresql_storage.Storage.schema_version = last_version
        self.storage.initialize_schema()

        # Check that the rotted tombstone has been removed, but the
        # original object remains.
        objects = self.storage.list_all("test", "jean-louis")
        # Only the object remains.
        assert len(objects) == 1

    def test_migration_18_merges_tombstones(self):
        last_version = postgresql_storage.Storage.schema_version

        self._load_schema("schema/postgresql-storage-11.sql")
        # Schema 11 is essentially the same as schema 17
        postgresql_storage.Storage.schema_version = 17
        with self.storage.client.connect() as conn:
            conn.execute(
                sa.text(
                    """
            UPDATE metadata SET value = '17'
            WHERE name = 'storage_schema_version';
            """
                )
            )

        insert_query = """
        INSERT INTO records (id, parent_id, collection_id, data, last_modified)
        VALUES (:id, :parent_id, :collection_id, (:data)::JSONB, from_epoch(:last_modified))
        """
        placeholders = dict(
            id="rid",
            parent_id="jean-louis",
            collection_id="test",
            data=json.dumps({"drink": "mate"}),
            last_modified=123456,
        )
        with self.storage.client.connect() as conn:
            conn.execute(sa.text(insert_query), placeholders)

        create_tombstone = """
        INSERT INTO deleted (id, parent_id, collection_id, last_modified)
        VALUES (:id, :parent_id, :collection_id, from_epoch(:last_modified))
        """
        with self.storage.client.connect() as conn:
            conn.execute(sa.text(create_tombstone), placeholders)

        # Execute the 017 to 018 migration (and others)
        postgresql_storage.Storage.schema_version = last_version
        self.storage.initialize_schema()

        # Check that the object took precedence of over the tombstone.
        objects = self.storage.list_all("test", "jean-louis", include_deleted=True)
        count = self.storage.count_all("test", "jean-louis")
        assert len(objects) == 1
        assert count == 1
        assert objects[0]["drink"] == "mate"


@skip_if_no_postgresql
class PostgresqlPermissionMigrationTest(unittest.TestCase):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        from kinto.core.utils import sqlalchemy

        if sqlalchemy is None:
            return

        from .test_permission import PostgreSQLPermissionTest

        settings = {**PostgreSQLPermissionTest.settings}
        config = testing.setUp()
        config.add_settings(settings)
        self.permission = postgresql_permission.load_from_config(config)

    def setUp(self):
        q = """
        DROP TABLE IF EXISTS access_control_entries CASCADE;
        DROP TABLE IF EXISTS user_principals CASCADE;
        DROP TABLE IF EXISTS metadata CASCADE;
        """
        with self.permission.client.connect() as conn:
            conn.execute(sa.text(q))

    def _load_schema(self, filepath):
        with self.permission.client.connect() as conn:
            here = os.path.dirname(__file__)
            with open(os.path.join(here, filepath)) as f:
                old_schema = f.read()
            conn.execute(sa.text(old_schema))

    def test_assumes_schema_1_if_no_permission_schema_version(self):
        self.permission.initialize_schema()
        q = """
        DELETE FROM metadata WHERE name = 'permission_schema_version';
        """
        with self.permission.client.connect() as conn:
            conn.execute(sa.text(q))

        self.assertEqual(self.permission.get_installed_version(), 1)

    def test_assumes_schema_1_if_no_metadata_table(self):
        self.permission.initialize_schema()
        q = """
        DROP TABLE metadata;
        """
        with self.permission.client.connect() as conn:
            conn.execute(sa.text(q))

        self.assertEqual(self.permission.get_installed_version(), 1)

    def test_assumes_no_schema_if_no_user_principals_table(self):
        # Verify that the permission migrations assume no schema if
        # the user_principals table is missing, even if the metadata
        # table is present from the storage migrations.
        self.permission.initialize_schema()
        q = """
        DELETE FROM metadata WHERE name = 'permission_schema_version';
        DROP TABLE user_principals;
        """
        with self.permission.client.connect() as conn:
            conn.execute(sa.text(q))

        self.assertEqual(self.permission.get_installed_version(), None)

    def test_schema_creation_leaves_at_max_version(self):
        self.permission.initialize_schema()
        self.assertEqual(self.permission.get_installed_version(), self.permission.schema_version)

    def test_runs_initialize_schema_if_using_it_fails(self):
        self.permission.initialize_schema()
        query = """SELECT 1 FROM information_schema.tables
        WHERE table_name = 'user_principals';"""
        with self.permission.client.connect(readonly=True) as conn:
            result = conn.execute(sa.text(query))
            self.assertEqual(result.rowcount, 1)

    def test_does_not_execute_if_ran_with_dry(self):
        self.permission.initialize_schema(dry_run=True)
        query = """SELECT 1 FROM information_schema.tables
        WHERE table_name = 'user_principals';"""
        with self.permission.client.connect(readonly=True) as conn:
            result = conn.execute(sa.text(query))
        self.assertEqual(result.rowcount, 0)

    def test_every_available_migration(self):
        """Test every permission migration available in code base.

        Objects migration test is currently very naive, and should be
        elaborated along future migrations.
        """
        # Install old schema
        self._load_schema("schema/postgresql-permission-1.sql")

        # Create some permissions data using some code that is
        # compatible with the schema in place.
        with self.permission.client.connect() as conn:
            query = """
            INSERT INTO user_principals
            VALUES ('remy', 'admin');
            INSERT INTO access_control_entries VALUES
              ('sailboat', 'write', 'remy'),
              ('sailboat', 'read', 'ethan'),
              ('sailboat/log', 'read', 'system.Authenticated'),
              ('sailboat/log', 'write', 'admin');
            """
            conn.execute(sa.text(query))

        version = self.permission.get_installed_version()
        self.assertEqual(version, 1)

        # Run every migrations available.
        self.permission.initialize_schema()

        # Version matches current one.
        version = self.permission.get_installed_version()
        self.assertEqual(version, self.permission.schema_version)

        # Check that previously created data is still available
        remy_principals = self.permission.get_user_principals("remy")
        self.assertEqual(remy_principals, set(["admin"]))

        remy_objects = self.permission.get_accessible_objects(
            ["remy", "admin", "system.Authenticated"]
        )
        self.assertEqual(
            remy_objects, {"sailboat": set(["write"]), "sailboat/log": set(["read", "write"])}
        )

        # Check that new objects can be created
        self.permission.add_user_principal("ethan", "crew")

        # And deleted
        self.permission.remove_principal_from_ace("sailboat/log", "read", "system.Authenticated")


@skip_if_no_postgresql
class PostgresqlCacheMigrationTest(unittest.TestCase):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        from kinto.core.utils import sqlalchemy

        if sqlalchemy is None:
            return

        from .test_cache import PostgreSQLCacheTest

        settings = {**PostgreSQLCacheTest.settings}
        config = testing.setUp()
        config.add_settings(settings)
        self.cache = postgresql_cache.load_from_config(config)

    def setUp(self):
        q = """
        DROP TABLE IF EXISTS cache CASCADE;
        """
        with self.cache.client.connect() as conn:
            conn.execute(sa.text(q))

    def test_runs_initialize_schema_if_using_it_fails(self):
        self.cache.initialize_schema()
        query = """SELECT 1 FROM information_schema.tables
        WHERE table_name = 'cache';"""
        with self.cache.client.connect(readonly=True) as conn:
            result = conn.execute(sa.text(query))
            self.assertEqual(result.rowcount, 1)

    def test_does_not_execute_if_ran_with_dry(self):
        self.cache.initialize_schema(dry_run=True)
        query = """SELECT 1 FROM information_schema.tables
        WHERE table_name = 'cache';"""
        with self.cache.client.connect(readonly=True) as conn:
            result = conn.execute(sa.text(query))
        self.assertEqual(result.rowcount, 0)


class PostgresqlExceptionRaisedTest(unittest.TestCase):
    def setUp(self):
        self.sqlalchemy = postgresql_storage.client.sqlalchemy

    def tearDown(self):
        postgresql_storage.client.sqlalchemy = self.sqlalchemy

    def test_postgresql_usage_raise_an_error_if_postgresql_not_installed(self):
        postgresql_storage.client.sqlalchemy = None
        with self.assertRaises(ImportWarning):
            postgresql_storage.client.create_from_config(testing.setUp())



import colander
from jsonschema import Draft7Validator as DraftValidator
from jsonschema import RefResolutionError, SchemaError, ValidationError
from jsonschema.validators import validator_for
from pyramid.settings import asbool

from kinto.core import utils
from kinto.core.errors import raise_invalid
from kinto.views import object_exists_or_404


class JSONSchemaMapping(colander.SchemaNode):
    def schema_type(self, **kw):
        return colander.Mapping(unknown="preserve")

    def deserialize(self, cstruct=colander.null):
        # Start by deserializing a simple mapping.
        validated = super().deserialize(cstruct)
        try:
            check_schema(validated)
        except ValidationError as e:
            self.raise_invalid(e.message)
        return validated


def check_schema(data):
    try:
        DraftValidator.check_schema(data)
    except SchemaError as e:
        message = e.path.pop() + e.message
        raise ValidationError(message)


# Module level global that stores a version of every possible schema (as a <class 'dict'>)
# turned into a jsonschema instance (as <class 'jsonschema.validators.Validator'>).
_schema_cache = {}


def validate(data, schema):
    """Raise a ValidationError or a RefResolutionError if the data doesn't validate
    with the given schema.

    Note that this function is just a "wrapper" on `jsonschema.validate()` but with
    some memoization based on the schema for better repeat performance.
    """
    # Because the schema is a dict, it can't be used as a hash key so it needs to be
    # "transformed" to something that is hashable. The quickest solution is to convert
    # it to a string.
    # Note that the order of the dict will determine the string it becomes. The solution
    # to that would a canonical serializer like `json.dumps(..., sort_keys=True)` but it's
    # overkill since the assumption is that the schema is very unlikely to be exactly
    # the same but different order.
    cache_key = str(schema)
    if cache_key not in _schema_cache:
        # This is essentially what the `jsonschema.validate()` shortcut function does.
        cls = validator_for(schema)
        cls.check_schema(schema)
        _schema_cache[cache_key] = cls(schema)
    return _schema_cache[cache_key].validate(data)


def validate_schema(data, schema, id_field, ignore_fields=None):
    if ignore_fields is None:
        ignore_fields = []
    # Only ignore the `id` field if the schema does not explicitly mention it.
    if id_field not in schema.get("properties", {}):
        ignore_fields += (id_field,)

    required_fields = [f for f in schema.get("required", []) if f not in ignore_fields]
    # jsonschema doesn't accept 'required': [] yet.
    # See https://github.com/Julian/jsonschema/issues/337.
    # In the meantime, strip out 'required' if no other fields are required.
    if required_fields:
        schema = {**schema, "required": required_fields}
    else:
        schema = {f: v for f, v in schema.items() if f != "required"}

    data = {f: v for f, v in data.items() if f not in ignore_fields}

    try:
        validate(data, schema)
    except ValidationError as e:
        if e.path:
            field = e.path[-1]
        elif e.validator_value:
            field = e.validator_value[-1]
        else:
            field = e.schema_path[-1]
        e.field = field
        raise e
    # Raise an error here if a reference in the schema doesn't resolve.
    # jsonschema doesn't provide schema validation checking upon creation yet,
    # it must be validated against data.
    # See https://github.com/Julian/jsonschema/issues/399
    # For future support https://github.com/Julian/jsonschema/issues/346.
    except RefResolutionError as e:
        raise e


def validate_from_bucket_schema_or_400(data, resource_name, request, id_field, ignore_fields):
    """Lookup in the parent objects if a schema was defined for this resource.

    If the schema validation feature is enabled, if a schema is/are defined, and if the
    data does not validate it/them, then it raises a 400 exception.
    """
    settings = request.registry.settings
    schema_validation = "experimental_collection_schema_validation"
    # If disabled from settings, do nothing.
    if not asbool(settings.get(schema_validation)):
        return

    bucket_id = request.matchdict["bucket_id"]
    bucket_uri = utils.instance_uri(request, "bucket", id=bucket_id)
    buckets = request.bound_data.setdefault("buckets", {})
    if bucket_uri not in buckets:
        # Unknown yet, fetch from storage.
        bucket = object_exists_or_404(
            request, resource_name="bucket", parent_id="", object_id=bucket_id
        )
        buckets[bucket_uri] = bucket

    # Let's see if the bucket defines a schema for this resource.
    metadata_field = f"{resource_name}:schema"
    bucket = buckets[bucket_uri]
    if metadata_field not in bucket:
        return

    # Validate or fail with 400.
    schema = bucket[metadata_field]
    try:
        validate_schema(data, schema, ignore_fields=ignore_fields, id_field=id_field)
    except ValidationError as e:
        raise_invalid(request, name=e.field, description=e.message)
    except RefResolutionError as e:
        raise_invalid(request, name="schema", description=str(e))



import unittest

from pyramid.authorization import Authenticated

from kinto.core.testing import get_user_headers

from .support import (
    MINIMALIST_BUCKET,
    MINIMALIST_COLLECTION,
    MINIMALIST_GROUP,
    MINIMALIST_RECORD,
    BaseWebTest,
)


class BucketViewTest(BaseWebTest, unittest.TestCase):
    collection_url = "/buckets"
    record_url = "/buckets/beers"

    def setUp(self):
        super().setUp()
        resp = self.app.put_json(self.record_url, MINIMALIST_BUCKET, headers=self.headers)
        self.record = resp.json["data"]

    def test_buckets_are_global_to_every_users(self):
        self.app.patch_json(
            self.record_url, {"permissions": {"read": [Authenticated]}}, headers=self.headers
        )
        self.app.get(self.record_url, headers=get_user_headers("alice"))

    def test_buckets_can_be_put_with_simple_name(self):
        self.assertEqual(self.record["id"], "beers")

    def test_buckets_names_can_have_underscores(self):
        bucket = {**MINIMALIST_BUCKET}
        record_url = "/buckets/alexis_beers"
        resp = self.app.put_json(record_url, bucket, headers=self.headers)
        self.assertEqual(resp.json["data"]["id"], "alexis_beers")

    def test_can_list_buckets_by_default_since_allowed_to_create(self):
        resp = self.app.get("/buckets", headers=get_user_headers("alice"), status=200)
        assert resp.json["data"] == []

    def test_buckets_name_should_be_simple(self):
        self.app.put_json(
            "/buckets/__beers__", MINIMALIST_BUCKET, headers=self.headers, status=400
        )

    def test_buckets_should_reject_unaccepted_request_content_type(self):
        headers = {**self.headers, "Content-Type": "text/plain"}
        self.app.put("/buckets/beers", MINIMALIST_BUCKET, headers=headers, status=415)

    def test_create_permissions_can_be_added_on_buckets(self):
        bucket = {
            **MINIMALIST_BUCKET,
            "permissions": {"collection:create": ["fxa:user"], "group:create": ["fxa:user"]},
        }
        resp = self.app.put_json("/buckets/beers", bucket, headers=self.headers, status=200)
        permissions = resp.json["permissions"]
        self.assertIn("fxa:user", permissions["collection:create"])
        self.assertIn("fxa:user", permissions["group:create"])

    def test_wrong_create_permissions_cannot_be_added_on_buckets(self):
        bucket = {**MINIMALIST_BUCKET, "permissions": {"record:create": ["fxa:user"]}}
        self.app.put_json("/buckets/beers", bucket, headers=self.headers, status=400)

    def test_buckets_can_handle_arbitrary_attributes(self):
        public_key = "5866f245a00bb3a39100d31b2f14d453"
        bucket = {**MINIMALIST_BUCKET, "data": {"public_key": public_key}}
        resp = self.app.put_json("/buckets/beers", bucket, headers=self.headers, status=200)
        data = resp.json["data"]
        self.assertIn("public_key", data)
        self.assertEqual(data["public_key"], public_key)

    def test_buckets_can_be_filtered_by_arbitrary_attribute(self):
        bucket = {**MINIMALIST_BUCKET, "data": {"size": 3}}
        self.app.put_json("/buckets/beers", bucket, headers=self.headers)
        resp = self.app.get("/buckets?min_size=2", headers=self.headers)
        data = resp.json["data"]
        self.assertEqual(len(data), 1)


class BucketListTest(BaseWebTest, unittest.TestCase):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["bucket_create_principals"] = cls.principal
        return settings

    def test_can_list_buckets_by_default_if_allowed_to_create(self):
        resp = self.app.get("/buckets", headers=self.headers, status=200)
        assert resp.json["data"] == []

    def test_cannot_list_buckets_if_empty_and_not_allowed_to_create(self):
        self.app.get("/buckets", status=401)
        self.app.get("/buckets", headers=get_user_headers("alice"), status=403)

    def test_can_list_buckets_if_some_are_shared(self):
        self.app.put_json("/buckets/whiskies", headers=self.headers)
        self.app.put_json(
            "/buckets/beers", {"permissions": {"write": ["system.Everyone"]}}, headers=self.headers
        )

        resp = self.app.get("/buckets", status=200)
        assert len(resp.json["data"]) == 1
        assert resp.json["data"][0]["id"] == "beers"


class BucketCreationTest(BaseWebTest, unittest.TestCase):
    def test_buckets_can_be_created_with_post(self):
        r = self.app.post_json("/buckets", MINIMALIST_BUCKET, headers=self.headers)
        self.assertEqual(r.status_code, 201)

    def test_bucket_id_can_be_specified_in_post(self):
        bucket = "blog"
        r = self.app.post_json("/buckets", {"data": {"id": bucket}}, headers=self.headers)
        self.assertEqual(r.json["data"]["id"], bucket)

    def test_bucket_can_be_created_without_body_nor_contenttype(self):
        headers = {**self.headers}
        headers.pop("Content-Type")
        self.app.put("/buckets/catalog", headers=headers)


class BucketReadPermissionTest(BaseWebTest, unittest.TestCase):
    collection_url = "/buckets"
    record_url = "/buckets/beers"

    def setUp(self):
        super().setUp()
        bucket = {**MINIMALIST_BUCKET}
        self.app.put_json(self.record_url, bucket, headers=self.headers)

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        # Give the right to list buckets (for cls.principal and alice).
        settings["kinto.bucket_read_principals"] = Authenticated
        return settings

    def test_bucket_collection_endpoint_lists_them_all_for_everyone(self):
        resp = self.app.get(self.collection_url, headers=get_user_headers("alice"))
        records = resp.json["data"]
        self.assertEqual(len(records), 1)
        self.assertEqual(records[0]["id"], "beers")

    def test_everyone_can_read_bucket_information(self):
        resp = self.app.get(self.record_url, headers=get_user_headers("alice"))
        record = resp.json["data"]
        self.assertEqual(record["id"], "beers")


class BucketDeletionTest(BaseWebTest, unittest.TestCase):
    bucket_url = "/buckets/beers"
    collection_url = "/buckets/beers/collections/barley"
    group_url = "/buckets/beers/groups/moderators"

    def setUp(self):
        # Create a bucket with some objects.
        self.app.put_json(self.bucket_url, MINIMALIST_BUCKET, headers=self.headers)
        self.app.put_json(self.group_url, MINIMALIST_GROUP, headers=self.headers)
        self.app.put_json(self.collection_url, MINIMALIST_COLLECTION, headers=self.headers)
        resp = self.app.post_json(
            self.collection_url + "/records", MINIMALIST_RECORD, headers=self.headers
        )
        record_id = resp.json["data"]["id"]
        self.previous_ts = resp.headers["ETag"]
        self.record_url = self.collection_url + "/records/{}".format(record_id)
        # Delete the bucket.
        self.app.delete(self.bucket_url, headers=self.headers)

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        # Give the permission to read, to get an explicit 404 once deleted.
        settings["kinto.bucket_read_principals"] = cls.principal
        return settings

    def test_buckets_can_be_deleted_in_bulk(self):
        self.app.put_json("/buckets/1", MINIMALIST_BUCKET, headers=get_user_headers("alice"))
        self.app.put_json("/buckets/2", MINIMALIST_BUCKET, headers=self.headers)
        self.app.put_json("/buckets/3", MINIMALIST_BUCKET, headers=self.headers)
        self.app.delete("/buckets", headers=self.headers)
        self.app.get("/buckets/1", headers=self.headers, status=200)
        self.app.get("/buckets/2", headers=self.headers, status=404)
        self.app.get("/buckets/3", headers=self.headers, status=404)

    def test_unknown_bucket_raises_404(self):
        resp = self.app.get("/buckets/2", headers=self.headers, status=404)
        self.assertEqual(resp.json["details"]["id"], "2")
        self.assertEqual(resp.json["details"]["resource_name"], "bucket")

    def test_buckets_can_be_deleted(self):
        self.app.get(self.bucket_url, headers=self.headers, status=404)

    def test_every_collections_are_deleted_too(self):
        self.app.put_json(self.bucket_url, MINIMALIST_BUCKET, headers=self.headers)
        self.app.get(self.collection_url, headers=self.headers, status=404)

        # Verify tombstones
        resp = self.app.get(
            "{}/collections?_since=0".format(self.bucket_url), headers=self.headers
        )
        self.assertEqual(len(resp.json["data"]), 0)

    def test_timestamps_are_refreshed_when_collection_is_recreated(self):
        # Kinto/kinto#1223
        # Recreate with same name.
        self.app.put_json(self.bucket_url, MINIMALIST_BUCKET, headers=self.headers)
        self.app.put_json(self.collection_url, MINIMALIST_COLLECTION, headers=self.headers)
        resp = self.app.get(self.collection_url + "/records", headers=self.headers)
        records_ts = resp.headers["ETag"]
        self.assertNotEqual(self.previous_ts, records_ts)

    def test_every_groups_are_deleted_too(self):
        self.app.put_json(self.bucket_url, MINIMALIST_BUCKET, headers=self.headers)
        self.app.get(self.group_url, headers=self.headers, status=404)
        # Verify tombstones
        resp = self.app.get("{}/groups?_since=0".format(self.bucket_url), headers=self.headers)
        self.assertEqual(len(resp.json["data"]), 0)

    def test_every_records_are_deleted_too(self):
        self.app.put_json(self.bucket_url, MINIMALIST_BUCKET, headers=self.headers)
        self.app.put_json(self.collection_url, MINIMALIST_COLLECTION, headers=self.headers)
        self.app.get(self.record_url, headers=self.headers, status=404)

        # Verify tombstones
        resp = self.app.get(
            "{}/records?_since=0".format(self.collection_url), headers=self.headers
        )
        self.assertEqual(len(resp.json["data"]), 0)

    def test_can_be_created_after_deletion_with_if_none_match_star(self):
        headers = {**self.headers, "If-None-Match": "*"}
        self.app.put_json(self.bucket_url, MINIMALIST_BUCKET, headers=headers, status=201)

    def test_does_not_delete_buckets_with_similar_names(self):
        self.app.put("/buckets/a", headers=self.headers)
        body = {"permissions": {"read": ["system.Everyone"]}}
        resp = self.app.put_json("/buckets/ab", body, headers=self.headers)
        before = resp.json

        self.app.delete("/buckets/a", headers=self.headers)

        resp = self.app.get("/buckets/ab", headers=self.headers, status=200)
        self.assertEqual(resp.json, before)



"""
A helper class to run migrations using a series of SQL files.
"""

import logging
import os

from kinto.core.utils import sqlalchemy as sa


logger = logging.getLogger(__name__)


class MigratorMixin:
    """Mixin to allow the running of migrations.

    Your class must provide a `client` attribute (a PostgreSQLClient),
    as well as override some class attributes.
    """

    """Name of this migrator (e.g. "storage"). Override this."""
    name = None

    """The current "newest" schema version. Override this."""
    schema_version = None

    """The file to find the current "newest" schema in. Override this."""
    schema_file = None

    """The directory to find migration files in.

    Migrations should each be a file named migration_nnn_mmm.sql, where mmm = nnn + 1.
    """
    migrations_directory = None

    def get_installed_version(self):
        """Return current version of schema or None if none found.

        Override this.

        This may be called several times during a single migration.
        """
        raise NotImplementedError("method not overridden")  # pragma: no cover

    def create_or_migrate_schema(self, dry_run=False):
        """Either create or migrate the schema, as needed."""
        version = self.get_installed_version()
        if not version:
            self.create_schema(dry_run)
            return

        logger.info(f"Detected PostgreSQL {self.name} schema version {version}.")
        if version == self.schema_version:
            logger.info(f"PostgreSQL {self.name} schema is up-to-date.")
            return

        self.migrate_schema(version, dry_run)

    def create_schema(self, dry_run):
        """Actually create the schema from scratch using self.schema_file.

        You can override this if you want to add additional sanity checks.
        """
        logger.info(
            f"Create PostgreSQL {self.name} schema at version {self.schema_version} from {self.schema_file}."
        )
        if not dry_run:
            self._execute_sql_file(self.schema_file)
            logger.info(f"Created PostgreSQL {self.name} schema (version {self.schema_version}).")

    def migrate_schema(self, start_version, dry_run):
        migrations = [(v, v + 1) for v in range(start_version, self.schema_version)]
        for migration in migrations:
            expected = migration[0]
            current = self.get_installed_version()
            error_msg = f"PostgreSQL {self.name} schema: Expected version {expected}. Found version {current}."
            if not dry_run and expected != current:
                raise AssertionError(error_msg)

            logger.info(
                f"Migrate PostgreSQL {self.name} schema from version {migration[0]} to {migration[1]}."
            )
            filename = "migration_{0:03d}_{1:03d}.sql".format(*migration)
            filepath = os.path.join(self.migrations_directory, filename)
            logger.info(f"Execute PostgreSQL {self.name} migration from {filepath}")
            if not dry_run:
                self._execute_sql_file(filepath)
        logger.info(
            f"PostgreSQL {self.name} schema migration {'simulated' if dry_run else 'done'}"
        )

    def _execute_sql_file(self, filepath):
        """Helper method to execute the SQL in a file."""
        with open(filepath) as f:
            schema = f.read()
        # Since called outside request, force commit.
        with self.client.connect(force_commit=True) as conn:
            conn.execute(sa.text(schema))



from kinto.core.testing import unittest

from .support import MINIMALIST_BUCKET, MINIMALIST_COLLECTION, MINIMALIST_RECORD, BaseWebTest


class GlobalSettingsTest(BaseWebTest, unittest.TestCase):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["kinto.record_cache_expires_seconds"] = 3600
        settings["kinto.record_read_principals"] = "system.Everyone"
        return settings

    def setUp(self):
        super().setUp()
        self.create_bucket("blog")
        self.app.put_json(
            "/buckets/blog/collections/cached", MINIMALIST_COLLECTION, headers=self.headers
        )
        r = self.app.post_json(
            "/buckets/blog/collections/cached/records", MINIMALIST_RECORD, headers=self.headers
        )
        self.record = r.json["data"]

    def test_expires_and_cache_control_headers_are_set(self):
        url = "/buckets/blog/collections/cached/records"
        r = self.app.get(url)
        self.assertIn("Expires", r.headers)
        self.assertEqual(r.headers["Cache-Control"], "max-age=3600")

        r = self.app.get("{}/{}".format(url, self.record["id"]))
        self.assertIn("Expires", r.headers)
        self.assertEqual(r.headers["Cache-Control"], "max-age=3600")


class SpecificSettingsTest(BaseWebTest, unittest.TestCase):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["kinto.blog.record_cache_expires_seconds"] = "30"
        settings["kinto.browser.top500.record_cache_expires_seconds"] = "60"
        return settings

    def setUp(self):
        super().setUp()

        def create_record_in_collection(bucket_id, collection_id):
            bucket = {**MINIMALIST_BUCKET, "permissions": {"read": ["system.Everyone"]}}
            self.app.put_json("/buckets/{}".format(bucket_id), bucket, headers=self.headers)
            collection_url = "/buckets/{}/collections/{}".format(bucket_id, collection_id)
            self.app.put_json(collection_url, MINIMALIST_COLLECTION, headers=self.headers)
            r = self.app.post_json(
                collection_url + "/records", MINIMALIST_RECORD, headers=self.headers
            )
            return r.json["data"]

        self.blog_record = create_record_in_collection("blog", "cached")
        self.app_record = create_record_in_collection("browser", "top500")

    def assertHasCache(self, url, age):
        r = self.app.get(url)
        self.assertIn("Expires", r.headers)
        self.assertEqual(r.headers["Cache-Control"], "max-age={}".format(age))

    def test_for_records_on_a_specific_bucket(self):
        collection_url = "/buckets/blog/collections/cached/records"
        self.assertHasCache(collection_url, 30)
        record_url = "{}/{}".format(collection_url, self.blog_record["id"])
        self.assertHasCache(record_url, 30)

    def test_for_records_on_a_specific_collection(self):
        collection_url = "/buckets/browser/collections/top500/records"
        self.assertHasCache(collection_url, 60)
        record_url = "{}/{}".format(collection_url, self.app_record["id"])
        self.assertHasCache(record_url, 60)


class CollectionExpiresTest(BaseWebTest, unittest.TestCase):
    def setUp(self):
        super().setUp()
        bucket = {**MINIMALIST_BUCKET, "permissions": {"read": ["system.Everyone"]}}
        self.app.put_json("/buckets/blog", bucket, headers=self.headers)
        self.collection_url = "/buckets/blog/collections/cached"
        self.app.put_json(
            self.collection_url, {"data": {"cache_expires": 3600}}, headers=self.headers
        )

        self.records_url = self.collection_url + "/records"

        resp = self.app.post_json(self.records_url, MINIMALIST_RECORD, headers=self.headers)
        self.record = resp.json["data"]
        self.record_url = "{}/{}".format(self.records_url, self.record["id"])

    def test_cache_expires_must_be_an_integer(self):
        self.app.put_json(
            self.collection_url,
            {"data": {"cache_expires": "abc"}},
            headers=self.headers,
            status=400,
        )

    def test_expires_and_cache_control_are_not_set_if_authenticated(self):
        r = self.app.get(self.records_url, headers=self.headers)
        self.assertNotIn("Expires", r.headers)
        self.assertIn("Cache-Control", r.headers)
        self.assertNotIn("max-age", r.headers["Cache-Control"])

    def test_expires_and_cache_control_are_set_on_records(self):
        r = self.app.get(self.records_url)
        self.assertIn("Expires", r.headers)
        self.assertEqual(r.headers["Cache-Control"], "max-age=3600")

    def test_expires_and_cache_control_are_set_on_record(self):
        r = self.app.get(self.record_url)
        self.assertIn("Expires", r.headers)
        self.assertEqual(r.headers["Cache-Control"], "max-age=3600")

    def test_cache_control_is_set_no_cache_if_zero(self):
        self.app.put_json(
            self.collection_url, {"data": {"cache_expires": 0}}, headers=self.headers
        )
        r = self.app.get(self.records_url)
        self.assertIn("Expires", r.headers)
        self.assertEqual(
            r.headers["Cache-Control"], "max-age=0, must-revalidate, no-cache, no-store"
        )
        self.assertEqual(r.headers["Pragma"], "no-cache")

    def test_cache_control_on_collection_overrides_setting(self):
        app = self.make_app(
            settings={
                "kinto.record_cache_expires_seconds": 10,
                "kinto.record_read_principals": "system.Everyone",
            }
        )
        app.put_json("/buckets/blog", MINIMALIST_BUCKET, headers=self.headers)
        app.put_json(self.collection_url, {"data": {"cache_expires": 3600}}, headers=self.headers)
        r = app.get(self.records_url)
        self.assertIn("Expires", r.headers)
        self.assertEqual(r.headers["Cache-Control"], "max-age=3600")



import logging
from enum import Enum

import colander
from pyramid import httpexceptions

from kinto.core.schema import Any
from kinto.core.utils import json, reapply_cors


class ERRORS(Enum):
    """Predefined errors as specified by the API.

    +-------------+-------+------------------------------------------------+
    | Status code | Errno | Description                                    |
    +=============+=======+================================================+
    | 401         | 104   | Missing Authorization Token                    |
    +-------------+-------+------------------------------------------------+
    | 401         | 105   | Invalid Authorization Token                    |
    +-------------+-------+------------------------------------------------+
    | 400         | 106   | request body was not valid JSON                |
    +-------------+-------+------------------------------------------------+
    | 400         | 107   | invalid request parameter                      |
    +-------------+-------+------------------------------------------------+
    | 400         | 108   | missing request parameter                      |
    +-------------+-------+------------------------------------------------+
    | 400         | 109   | invalid posted data                            |
    +-------------+-------+------------------------------------------------+
    | 404         | 110   | Invalid Token / id                             |
    +-------------+-------+------------------------------------------------+
    | 404         | 111   | Missing Token / id                             |
    +-------------+-------+------------------------------------------------+
    | 411         | 112   | Content-Length header was not provided         |
    +-------------+-------+------------------------------------------------+
    | 413         | 113   | Request body too large                         |
    +-------------+-------+------------------------------------------------+
    | 412         | 114   | Resource was modified meanwhile                |
    +-------------+-------+------------------------------------------------+
    | 405         | 115   | Method not allowed on this end point           |
    +-------------+-------+------------------------------------------------+
    | 404         | 116   | Requested version not available on this server |
    +-------------+-------+------------------------------------------------+
    | 429         | 117   | Client has sent too many requests              |
    +-------------+-------+------------------------------------------------+
    | 403         | 121   | Resource's access forbidden for this user      |
    +-------------+-------+------------------------------------------------+
    | 409         | 122   | Another resource violates constraint           |
    +-------------+-------+------------------------------------------------+
    | 500         | 999   | Internal Server Error                          |
    +-------------+-------+------------------------------------------------+
    | 503         | 201   | Service Temporary unavailable due to high load |
    +-------------+-------+------------------------------------------------+
    | 410         | 202   | Service deprecated                             |
    +-------------+-------+------------------------------------------------+
    """

    MISSING_AUTH_TOKEN = 104
    INVALID_AUTH_TOKEN = 105
    BADJSON = 106
    INVALID_PARAMETERS = 107
    MISSING_PARAMETERS = 108
    INVALID_POSTED_DATA = 109
    INVALID_RESOURCE_ID = 110
    MISSING_RESOURCE = 111
    MISSING_CONTENT_LENGTH = 112
    REQUEST_TOO_LARGE = 113
    MODIFIED_MEANWHILE = 114
    METHOD_NOT_ALLOWED = 115
    VERSION_NOT_AVAILABLE = 116
    CLIENT_REACHED_CAPACITY = 117
    FORBIDDEN = 121
    CONSTRAINT_VIOLATED = 122
    UNDEFINED = 999
    BACKEND = 201
    SERVICE_DEPRECATED = 202


class ErrorSchema(colander.MappingSchema):
    """Payload schema for Kinto errors."""

    code = colander.SchemaNode(colander.Integer())
    errno = colander.SchemaNode(colander.Integer())
    error = colander.SchemaNode(colander.String())
    message = colander.SchemaNode(colander.String(), missing=colander.drop)
    info = colander.SchemaNode(colander.String(), missing=colander.drop)
    details = colander.SchemaNode(Any(), missing=colander.drop)


def http_error(
    httpexception, errno=None, code=None, error=None, message=None, info=None, details=None
):
    """Return a JSON formated response matching the error HTTP API.

    :param httpexception: Instance of :mod:`~pyramid:pyramid.httpexceptions`
    :param errno: stable application-level error number (e.g. 109)
    :param code: matches the HTTP status code (e.g 400)
    :param error: string description of error type (e.g. "Bad request")
    :param message: context information (e.g. "Invalid request parameters")
    :param info: information about error (e.g. URL to troubleshooting)
    :param details: additional structured details (conflicting object)
    :returns: the formatted response object
    :rtype: pyramid.httpexceptions.HTTPException
    """
    errno = errno or ERRORS.UNDEFINED

    if isinstance(errno, Enum):
        errno = errno.value

    body = {
        "code": code or httpexception.code,
        "errno": errno,
        "error": error or httpexception.title,
        "message": message,
        "info": info,
        "details": details or colander.drop,
    }

    response = httpexception
    response.errno = errno
    response.json = ErrorSchema().deserialize(body)
    response.content_type = "application/json"
    return response


def json_error_handler(request):
    """Cornice JSON error handler, returning consistant JSON formatted errors
    from schema validation errors.

    This is meant to be used is custom services in your applications.

    .. code-block:: python

        upload = Service(name="upload", path='/upload',
                         error_handler=errors.json_error_handler)

    .. warning::

        Only the first error of the list is formatted in the response.
        (c.f. HTTP API).
    """
    errors = request.errors
    sorted_errors = sorted(errors, key=lambda x: str(x["name"]))
    # In Cornice, we call error handler if at least one error was set.
    error = sorted_errors[0]
    name = error["name"]
    description = error["description"]

    if isinstance(description, bytes):
        description = error["description"].decode("utf-8")

    if name is not None:
        if str(name) in description:
            message = description
        else:
            message = "{name} in {location}: {description}".format_map(error)
    else:
        message = "{location}: {description}".format_map(error)

    response = http_error(
        httpexceptions.HTTPBadRequest(),
        code=errors.status,
        errno=ERRORS.INVALID_PARAMETERS.value,
        error="Invalid parameters",
        message=message,
        details=errors,
    )
    response.status = errors.status
    response = reapply_cors(request, response)
    return response


def raise_invalid(request, location="body", name=None, description=None, **kwargs):
    """Helper to raise a validation error.

    :param location: location in request (e.g. ``'querystring'``)
    :param name: field name
    :param description: detailed description of validation error

    :raises: :class:`~pyramid:pyramid.httpexceptions.HTTPBadRequest`
    """
    request.errors.add(location, name, description, **kwargs)
    response = json_error_handler(request)
    raise response


def send_alert(request, message=None, url=None, code="soft-eol"):
    """Helper to add an Alert header to the response.

    :param code: The type of error 'soft-eol', 'hard-eol'
    :param message: The description message.
    :param url: The URL for more information, default to the documentation url.
    """
    if url is None:
        url = request.registry.settings["project_docs"]

    request.response.headers["Alert"] = json.dumps({"code": code, "message": message, "url": url})


def request_GET(request):
    """Catches a UnicodeDecode error in request.GET in case a wrong request was received.
    Fixing a webob long term issue: https://github.com/Pylons/webob/issues/161
    """
    try:
        return request.GET
    except UnicodeDecodeError:
        querystring = request.environ.get("QUERY_STRING", "")
        logger = logging.getLogger(__name__)
        logger.info("Error decoding QUERY_STRING: %s" % request.environ)
        raise http_error(
            httpexceptions.HTTPBadRequest(),
            errno=ERRORS.INVALID_PARAMETERS,
            message="A request with an incorrect encoding in the querystring was"
            "received. Please make sure your requests are encoded in UTF-8: %s" % querystring,
        )



from pyramid import httpexceptions

from kinto.core.resource import Resource

from . import BaseTest


class PartialResponseBase(BaseTest):
    def setUp(self):
        super().setUp()
        self.resource._get_known_fields = lambda: ["field", "other", "orig"]
        self.obj = self.model.create_object(
            {
                "field": "value",
                "other": "val",
                "orig": {
                    "foo": "food",
                    "bar": "baz",
                    "nested": {"size": 12546, "hash": "0x1254", "mime": "image/png"},
                },
            }
        )
        self.resource.object_id = self.obj["id"]
        self.resource.request = self.get_request()
        self.resource.request.validated = self.validated
        self.validated["querystring"] = {}


class PartialFieldsTest(PartialResponseBase):
    def test_fields_parameter_do_projection_on_get(self):
        self.validated["querystring"]["_fields"] = ["field"]
        obj = self.resource.get()
        self.assertIn("field", obj["data"])
        self.assertNotIn("other", obj["data"])

    def test_fields_parameter_do_projection_on_get_all(self):
        self.validated["querystring"]["_fields"] = ["field"]
        obj = self.resource.plural_get()["data"][0]
        self.assertIn("field", obj)
        self.assertNotIn("other", obj)

    def test_fail_if_fields_parameter_is_invalid(self):
        self.validated["querystring"]["_fields"] = "invalid_field"
        self.assertRaises(httpexceptions.HTTPBadRequest, self.resource.get)
        self.assertRaises(httpexceptions.HTTPBadRequest, self.resource.plural_get)

    def test_can_have_multiple_fields(self):
        self.validated["querystring"]["_fields"] = ["field", "other"]
        obj = self.resource.get()
        self.assertIn("field", obj["data"])
        self.assertIn("other", obj["data"])

    def test_id_and_last_modified_are_not_filtered(self):
        self.validated["querystring"]["_fields"] = ["field"]
        obj = self.resource.get()
        self.assertIn("id", obj["data"])
        self.assertIn("last_modified", obj["data"])

    def test_nested_parameter_can_be_filtered(self):
        self.validated["querystring"]["_fields"] = ["orig.foo"]
        obj = self.resource.get()
        self.assertIn("orig", obj["data"])
        self.assertIn("foo", obj["data"]["orig"])
        self.assertNotIn("other", obj["data"])
        self.assertNotIn("bar", obj["data"]["orig"])
        self.assertNotIn("nested", obj["data"]["orig"])

    def test_nested_parameter_can_be_filtered_on_multiple_levels(self):
        self.validated["querystring"]["_fields"] = ["orig.nested.size"]
        obj = self.resource.get()
        self.assertIn("nested", obj["data"]["orig"])
        self.assertIn("size", obj["data"]["orig"]["nested"])
        self.assertNotIn("hash", obj["data"]["orig"]["nested"])
        self.assertNotIn("mime", obj["data"]["orig"]["nested"])

    def test_can_filter_on_several_nested_fields(self):
        self.validated["querystring"]["_fields"] = ["orig.nested.size", "orig.nested.hash"]
        obj = self.resource.get()
        self.assertIn("size", obj["data"]["orig"]["nested"])
        self.assertIn("hash", obj["data"]["orig"]["nested"])
        self.assertNotIn("mime", obj["data"]["orig"]["nested"])


class PermissionTest(PartialResponseBase):
    resource_class = Resource

    def test_permissions_are_not_displayed(self):
        self.validated["querystring"]["_fields"] = ["field"]
        result = self.resource.get()
        self.assertNotIn("permissions", result)



import unittest

from kinto.core.testing import get_user_headers

from .support import (
    MINIMALIST_BUCKET,
    MINIMALIST_COLLECTION,
    MINIMALIST_GROUP,
    MINIMALIST_RECORD,
    BaseWebTest,
)


class PermissionsTest(BaseWebTest, unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.alice_headers = {**cls.headers, **get_user_headers("alice")}
        cls.bob_headers = {**cls.headers, **get_user_headers("bob")}

        cls.alice_principal = (
            "basicauth:d5b0026601f1b251974e09548d44155e16812e3c64ff7ae053fe3542e2ca1570"
        )
        cls.bob_principal = (
            "basicauth:c031ced27503f788b102ca54269a062ec73794bb075154c74a0d4311e74ca8b6"
        )


class BucketPermissionsTest(PermissionsTest):
    def setUp(self):
        bucket = {**MINIMALIST_BUCKET, "permissions": {"read": [self.alice_principal]}}
        self.app.put_json("/buckets/sodas", bucket, headers=self.headers)

    def test_creation_is_allowed_to_authenticated_by_default(self):
        self.app.put_json("/buckets/beer", MINIMALIST_BUCKET, headers=self.headers)

    def test_current_user_receives_write_permission_on_creation(self):
        resp = self.app.put_json("/buckets/beer", MINIMALIST_BUCKET, headers=self.headers)
        permissions = resp.json["permissions"]
        self.assertIn(self.principal, permissions["write"])

    def test_can_read_if_allowed(self):
        self.app.get("/buckets/sodas", headers=self.alice_headers)

    def test_cannot_write_if_not_allowed(self):
        self.app.put_json(
            "/buckets/sodas", MINIMALIST_BUCKET, headers=self.alice_headers, status=403
        )

    def test_permissions_are_not_returned_if_can_only_read(self):
        resp = self.app.get("/buckets/sodas", headers=self.alice_headers)
        self.assertEqual(resp.json["permissions"], {})

    def test_permissions_are_returned_if_can_write(self):
        resp = self.app.get("/buckets/sodas", headers=self.headers)
        self.assertIn("write", resp.json["permissions"])

    def test_cannot_post_existing_id_if_cannot_read(self):
        self.app.get("/buckets/sodas", headers=self.bob_headers, status=403)
        self.app.post_json(
            "/buckets", {"data": {"id": "sodas"}}, headers=self.bob_headers, status=403
        )

    def test_can_post_existing_id_if_can_read(self):
        self.app.patch_json(
            "/buckets/sodas",
            {"data": {"marker": True}, "permissions": {"read": ["system.Authenticated"]}},
            headers=self.headers,
        )
        resp = self.app.post_json("/buckets", {"data": {"id": "sodas"}}, headers=self.bob_headers)
        assert resp.json["data"]["marker"]


class CollectionPermissionsTest(PermissionsTest):
    def setUp(self):
        bucket = {
            **MINIMALIST_BUCKET,
            "permissions": {"read": [self.alice_principal], "write": [self.bob_principal]},
        }
        self.app.put_json("/buckets/beer", bucket, headers=self.headers)
        self.app.put_json(
            "/buckets/beer/collections/barley",
            MINIMALIST_COLLECTION,
            headers=self.headers,
        )

    def test_passing_unicode_on_parent_id_is_supported(self):
        self.app.get(
            "/buckets/block%C2%93%C2%96sts/collections/barley",
            headers=self.alice_headers,
            status=403,
        )

    def test_read_is_allowed_if_read_on_bucket(self):
        self.app.get("/buckets/beer/collections/barley", headers=self.alice_headers)

    def test_read_is_allowed_if_write_on_bucket(self):
        self.app.get("/buckets/beer/collections/barley", headers=self.bob_headers)

    def test_cannot_read_if_not_allowed(self):
        headers = {**self.headers, **get_user_headers("jean-louis")}
        self.app.get("/buckets/beer/collections/barley", headers=headers, status=403)

    def test_cannot_write_if_not_allowed(self):
        self.app.put_json(
            "/buckets/beer/collections/barley",
            MINIMALIST_COLLECTION,
            headers=self.alice_headers,
            status=403,
        )

    def test_permission_backend_prevent_sql_injections(self):
        self.app.get("/buckets/beer'", headers=self.headers, status=403)
        self.app.get("/buckets/beer'/collections/barley", headers=self.headers, status=403)
        self.app.get("/buckets/beer'/groups/barley", headers=self.headers, status=403)

        self.app.get("/buckets/beer/collections/barley'", headers=self.headers, status=400)
        # XXX: We should validate the collection ID on the records collection endpoint. #1077
        self.app.get(
            "/buckets/beer/collections/barley'/records",
            headers=self.headers,
            status=404,
        )

        self.app.get("/buckets/beer/groups/barley'", headers=self.headers, status=400)


class GroupPermissionsTest(PermissionsTest):
    def setUp(self):
        bucket = {
            **MINIMALIST_BUCKET,
            "permissions": {"read": [self.alice_principal], "write": [self.bob_principal]},
        }
        self.app.put_json("/buckets/beer", bucket, headers=self.headers)

        self.app.put_json(
            "/buckets/beer/groups/moderators", MINIMALIST_GROUP, headers=self.headers
        )

    def test_creation_is_allowed_if_write_on_bucket(self):
        self.app.post_json("/buckets/beer/groups", MINIMALIST_GROUP, headers=self.headers)

    def test_read_is_allowed_if_read_on_bucket(self):
        self.app.get("/buckets/beer/groups/moderators", headers=self.alice_headers)

    def test_read_is_allowed_if_write_on_bucket(self):
        self.app.get("/buckets/beer/groups/moderators", headers=self.bob_headers)

    def test_cannot_read_if_not_allowed(self):
        headers = {**self.headers, **get_user_headers("jean-louis")}
        self.app.get("/buckets/beer/groups/moderators", headers=headers, status=403)

    def test_cannot_write_if_not_allowed(self):
        self.app.put_json(
            "/buckets/beer/groups/moderators",
            MINIMALIST_GROUP,
            headers=self.alice_headers,
            status=403,
        )

    def test_creation_is_forbidden_is_no_write_on_bucket(self):
        self.app.post_json(
            "/buckets/beer/groups",
            MINIMALIST_GROUP,
            headers=self.alice_headers,
            status=403,
        )


class RecordPermissionsTest(PermissionsTest):
    def setUp(self):
        bucket = {**MINIMALIST_BUCKET, "permissions": {"write": [self.alice_principal]}}
        self.app.put_json("/buckets/beer", bucket, headers=self.headers)

        collection = {
            **MINIMALIST_COLLECTION,
            "permissions": {"write": [self.bob_principal]},
        }
        self.app.put_json("/buckets/beer/collections/barley", collection, headers=self.headers)

    def test_creation_is_allowed_if_write_on_bucket(self):
        self.app.post_json(
            "/buckets/beer/collections/barley/records",
            MINIMALIST_RECORD,
            headers=self.alice_headers,
        )

    def test_creation_is_allowed_if_write_on_collection(self):
        self.app.post_json(
            "/buckets/beer/collections/barley/records",
            MINIMALIST_RECORD,
            headers=self.bob_headers,
        )

    def test_creation_is_forbidden_is_no_write_on_bucket_nor_collection(self):
        headers = {**self.headers, **get_user_headers("jean-louis")}
        self.app.post_json(
            "/buckets/beer/collections/barley/records",
            MINIMALIST_RECORD,
            headers=headers,
            status=403,
        )

    def test_record_permissions_are_modified_by_patch(self):
        collection_url = "/buckets/beer/collections/barley/records"
        resp = self.app.post_json(collection_url, MINIMALIST_RECORD, headers=self.headers)
        record = resp.json["data"]
        resp = self.app.patch_json(
            "{}/{}".format(collection_url, record["id"]),
            {"permissions": {"read": ["fxa:user"]}},
            headers=self.headers,
        )
        self.assertIn("fxa:user", resp.json["permissions"]["read"])


class ChildrenCreationTest(PermissionsTest):
    def setUp(self):
        self.app.put_json(
            "/buckets/create",
            {"permissions": {"group:create": ["system.Authenticated"]}},
            headers=self.alice_headers,
        )
        self.app.put_json(
            "/buckets/write",
            {"permissions": {"write": ["system.Authenticated"]}},
            headers=self.alice_headers,
        )
        self.app.put_json(
            "/buckets/read",
            {"permissions": {"read": ["system.Authenticated"]}},
            headers=self.alice_headers,
        )
        for parent in ("create", "write", "read"):
            self.app.put_json(
                "/buckets/{}/groups/child".format(parent),
                MINIMALIST_GROUP,
                headers=self.alice_headers,
            )
        self.bob_headers_safe_creation = dict({"If-None-Match": "*"}, **self.bob_headers)

    def test_cannot_read_others_objects_if_only_allowed_to_create(self):
        self.app.get("/buckets/create/groups/child", headers=self.bob_headers, status=403)

    def test_safe_creation_with_put_returns_412_if_allowed_to_create(self):
        self.app.put_json(
            "/buckets/create/groups/child",
            MINIMALIST_GROUP,
            headers=self.bob_headers_safe_creation,
            status=412,
        )

    def test_safe_creation_with_post_returns_412_if_allowed_to_create_and_read(self):
        self.app.patch_json(
            "/buckets/create/groups/child",
            {"permissions": {"read": ["system.Authenticated"]}},
            headers=self.alice_headers,
        )
        self.app.post_json(
            "/buckets/create/groups",
            {"data": {"id": "child", "members": []}},
            headers=self.bob_headers_safe_creation,
            status=412,
        )

    def test_safe_creation_with_put_returns_412_if_allowed_to_write(self):
        self.app.put_json(
            "/buckets/write/groups/child",
            MINIMALIST_GROUP,
            headers=self.bob_headers_safe_creation,
            status=412,
        )

    def test_safe_creation_with_post_returns_412_if_allowed_to_write(self):
        self.app.post_json(
            "/buckets/write/groups",
            {"data": {"id": "child", "members": []}},
            headers=self.bob_headers_safe_creation,
            status=412,
        )

    def test_safe_creation_with_put_returns_403_if_only_allowed_to_read(self):
        self.app.put_json(
            "/buckets/read/groups/child",
            MINIMALIST_GROUP,
            headers=self.bob_headers_safe_creation,
            status=403,
        )

    def test_safe_creation_with_post_returns_403_if_not_allowed_to_read(self):
        self.app.post_json(
            "/buckets/create/groups",
            {"data": {"id": "child", "members": []}},
            headers=self.bob_headers_safe_creation,
            status=403,
        )

    def test_safe_creation_with_post_returns_412_if_only_allowed_to_read(self):
        self.app.post_json(
            "/buckets/read/groups",
            {"data": {"id": "child", "members": []}},
            headers=self.bob_headers_safe_creation,
            status=412,
        )

    def test_delete_returns_404_on_unknown_if_only_allowed_to_read(self):
        self.app.delete("/buckets/read/groups/g1", headers=self.bob_headers, status=404)

    def test_patch_returns_404_on_unknown_if_only_allowed_to_read(self):
        self.app.patch_json(
            "/buckets/read/groups/g1",
            {"data": {"members": []}},
            headers=self.bob_headers,
            status=404,
        )


class ParentMetadataTest(PermissionsTest):
    def setUp(self):
        self.app.put_json(
            "/buckets/beer",
            {"permissions": {"collection:create": [self.bob_principal]}},
            headers=self.headers,
        )

        self.app.put_json("/buckets/beer/collections/wheat", headers=self.headers)
        self.app.put_json("/buckets/beer/collections/root", headers=self.headers)

        self.app.put_json(
            "/buckets/beer/collections/barley",
            {"permissions": {"record:create": [self.alice_principal]}},
            headers=self.bob_headers,
        )

    def test_parent_metadata_can_be_read_if_allowed_to_create_child(self):
        self.app.get("/buckets/beer", headers=self.bob_headers)
        self.app.get("/buckets/beer/collections/barley", headers=self.alice_headers)

    def test_parent_metadata_cannot_be_read_if_not_allowed_to_create_child(self):
        self.app.get("/buckets/beer", headers=get_user_headers("jean:paul"), status=403)
        self.app.get(
            "/buckets/beer/collections/barley",
            headers=get_user_headers("mahmud:hatim"),
            status=403,
        )

    def test_list_can_be_obtained_if_allowed_to_create(self):
        resp = self.app.get("/buckets/beer/collections", headers=self.bob_headers)
        self.assertEqual(len(resp.json["data"]), 1)
        self.assertEqual(resp.json["data"][0]["id"], "barley")

        resp = self.app.get("/buckets/beer/collections/barley/records", headers=self.alice_headers)
        self.assertEqual(resp.json["data"], [])

    def test_list_is_denied_if_not_allowed_to_create(self):
        self.app.get(
            "/buckets/beer/collections",
            headers=get_user_headers("jean:paul"),
            status=403,
        )
        self.app.get(
            "/buckets/beer/collections/barley/records",
            headers=get_user_headers("mahmud:hatim"),
            status=403,
        )


class DisabledExplicitPermissionsTest(BaseWebTest, unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.alice_headers = {**cls.headers, **get_user_headers("alice")}
        cls.alice_principal = (
            "basicauth:d5b0026601f1b251974e09548d44155e16812e3c64ff7ae053fe3542e2ca1570"
        )

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["explicit_permissions"] = "false"
        settings["experimental_permissions_endpoint"] = "true"
        return settings

    def setUp(self):
        self.app.put_json(
            "/buckets/write",
            {"permissions": {"write": ["system.Authenticated"]}},
            headers=self.headers,
        )
        self.app.put_json(
            "/buckets/write/collections/test",
            {"permissions": {"write": ["system.Authenticated"]}},
            headers=self.alice_headers,
        )

    def test_can_create_and_access_child_object(self):
        self.app.put(
            "/buckets/write/collections/test/records/1",
            headers=self.alice_headers,
        )
        self.app.get(
            "/buckets/write/collections/test/records/1",
            headers=self.alice_headers,
        )

    def test_current_user_is_not_added_to_object_permissions(self):
        resp = self.app.put_json(
            "/buckets/write/collections/test/records/1",
            {"permissions": {"write": ["system.Authenticated"], "read": ["ldap:chantal"]}},
            headers=self.alice_headers,
        )
        self.assertEqual(
            resp.json["permissions"],
            {"write": ["system.Authenticated"], "read": ["ldap:chantal"]},
        )

    def test_child_objects_are_not_listed_in_permission_endpoint(self):
        self.app.put(
            "/buckets/write/collections/test/records/1",
            headers=self.alice_headers,
        )

        resp = self.app.get("/permissions", headers=self.alice_headers)
        perms = resp.json["data"]

        self.assertEqual(
            sorted(p["uri"] for p in perms),
            ["/", "/buckets/write", "/buckets/write/collections/test"],
        )

    def test_write_via_groups(self):
        self.app.put_json(
            "/buckets/viagroup",
            {"permissions": {"write": [self.principal]}},
            headers=self.headers,
        )
        self.app.put_json(
            "/buckets/viagroup/collections/c",
            {"permissions": {"write": ["/buckets/viagroup/groups/editors"]}},
            headers=self.headers,
        )
        self.app.put_json(
            "/buckets/viagroup/groups/editors",
            {"data": {"members": [self.alice_principal]}},
            headers=self.headers,
        )

        self.app.post_json(
            "/buckets/viagroup/collections/c/records",
            {},
            headers=self.alice_headers,
        )



import unittest

from kinto import main

from .support import BaseWebTest


class TestMain(BaseWebTest, unittest.TestCase):
    def test_init_sets_command_on_registry(self):
        app = main({"command": "migrate"}, None, **self.get_app_settings())

        self.assertEqual(app.registry.command, "migrate")

    def test_init_with_no_global_config_doesnt_crash(self):
        app = main(None, None, **self.get_app_settings())

        self.assertEqual(app.registry.command, None)



from pyramid.authorization import Authenticated
from pyramid.settings import asbool

from kinto.core import resource, utils
from kinto.core.errors import raise_invalid
from kinto.schema_validation import (
    RefResolutionError,
    ValidationError,
    validate_from_bucket_schema_or_400,
    validate_schema,
)
from kinto.views import object_exists_or_404


_parent_path = "/buckets/{{bucket_id}}/collections/{{collection_id}}"


@resource.register(
    name="record",
    plural_path=_parent_path + "/records",
    object_path=_parent_path + "/records/{{id}}",
)
class Record(resource.Resource):
    schema_field = "schema"

    def __init__(self, request, **kwargs):
        # Before all, first check that the parent collection exists.
        # Check if already fetched before (in batch).
        collections = request.bound_data.setdefault("collections", {})
        collection_uri = self.get_parent_id(request)
        if collection_uri not in collections:
            # Unknown yet, fetch from storage.
            bucket_uri = utils.instance_uri(request, "bucket", id=self.bucket_id)
            collection = object_exists_or_404(
                request,
                resource_name="collection",
                parent_id=bucket_uri,
                object_id=self.collection_id,
            )
            collections[collection_uri] = collection
        self._collection = collections[collection_uri]

        super().__init__(request, **kwargs)

    def get_parent_id(self, request):
        self.bucket_id = request.matchdict["bucket_id"]
        self.collection_id = request.matchdict["collection_id"]
        return utils.instance_uri(
            request, "collection", bucket_id=self.bucket_id, id=self.collection_id
        )

    def process_object(self, new, old=None):
        """Validate records against collection or bucket schema, if any."""
        new = super().process_object(new, old)

        # Is schema validation enabled?
        settings = self.request.registry.settings
        schema_validation = "experimental_collection_schema_validation"
        if not asbool(settings.get(schema_validation)):
            return new

        # Remove internal and auto-assigned fields from schemas and record.
        ignored_fields = (
            self.model.modified_field,
            self.schema_field,
            self.model.permissions_field,
        )

        # The schema defined on the collection will be validated first.
        if "schema" in self._collection:
            schema = self._collection["schema"]
            try:
                validate_schema(
                    new, schema, ignore_fields=ignored_fields, id_field=self.model.id_field
                )
            except ValidationError as e:
                raise_invalid(self.request, name=e.field, description=e.message)
            except RefResolutionError as e:
                raise_invalid(self.request, name="schema", description=str(e))

            # Assign the schema version to the record.
            schema_timestamp = self._collection[self.model.modified_field]
            new[self.schema_field] = schema_timestamp

        # Validate also from the record:schema field defined on the bucket.
        validate_from_bucket_schema_or_400(
            new,
            resource_name="record",
            request=self.request,
            ignore_fields=ignored_fields,
            id_field=self.model.id_field,
        )

        return new

    def plural_get(self):
        result = super().plural_get()
        self._handle_cache_expires(self.request.response)
        return result

    def get(self):
        result = super().get()
        self._handle_cache_expires(self.request.response)
        return result

    def _handle_cache_expires(self, response):
        """If the parent collection defines a ``cache_expires`` attribute,
        then cache-control response headers are sent.

        .. note::

            Those headers are also sent if the
            ``kinto.record_cache_expires_seconds`` setting is defined.
        """
        is_anonymous = Authenticated not in self.request.effective_principals
        if not is_anonymous:
            return

        cache_expires = self._collection.get("cache_expires")
        if cache_expires is None:
            by_collection = f"{self.bucket_id}.{self.collection_id}.record_cache_expires_seconds"
            by_bucket = f"{self.bucket_id}.record_cache_expires_seconds"
            by_collection_legacy = (
                f"{self.bucket_id}_{self.collection_id}_record_cache_expires_seconds"
            )
            by_bucket_legacy = f"{self.bucket_id}_record_cache_expires_seconds"
            settings = self.request.registry.settings
            for s in (by_collection, by_bucket, by_collection_legacy, by_bucket_legacy):
                cache_expires = settings.get(s)
                if cache_expires is not None:
                    break
        if cache_expires is not None:
            response.cache_expires(seconds=int(cache_expires))



import codecs
import os
import tempfile
import unittest
from datetime import datetime, timedelta
from time import strftime
from unittest import mock

from kinto import __version__, config


class ConfigTest(unittest.TestCase):
    def _assertTimestampStringsAlmostEqual(self, s1, s2, delta=timedelta(seconds=1)):
        """Assert that two timestamp strings are almost equal, within a
        specified timedelta.

        :param s1: a string representing a timestamp with the format
            format %a, %d %b %Y %H:%M:%S %z
        :param s2: a string representing a timestamp with the format
            format %a, %d %b %Y %H:%M:%S %z
        :param delta: timedelta, default is 1 second
        :returns: True, if time between s1 and s2 is less than delta.
            False, otherwise.

        """
        s1 = datetime.strptime(s1, "%a, %d %b %Y %H:%M:%S %z")
        s2 = datetime.strptime(s2, "%a, %d %b %Y %H:%M:%S %z")

        return self.assertLessEqual(
            abs(s1 - s2),
            delta,
            f"Delta between {s1} and {s2} is {s1 - s2}. Expected difference "
            f"to be less than or equal to {delta}",
        )

    def test_transpose_parameters_into_template(self):
        self.maxDiff = None
        template = "kinto.tpl"
        dest = tempfile.mktemp()
        config.render_template(
            template,
            dest,
            host="127.0.0.1",
            secret="secret",
            bucket_id_salt="bucket_id_salt",
            storage_backend="storage_backend",
            cache_backend="cache_backend",
            permission_backend="permission_backend",
            storage_url="storage_url",
            cache_url="cache_url",
            permission_url="permission_url",
            kinto_version="kinto_version",
            config_file_timestamp="config_file_timestamp",
        )

        with codecs.open(dest, "r", encoding="utf-8") as d:
            destination_temp = d.read()

        sample_path = os.path.join(os.path.dirname(__file__), "test_configuration/test.ini")
        with codecs.open(sample_path, "r", encoding="utf-8") as c:
            sample = c.read()

        self.assertEqual(destination_temp, sample)

    def test_create_destination_directory(self):
        dest = os.path.join(tempfile.mkdtemp(), "config", "kinto.ini")

        config.render_template(
            "kinto.tpl",
            dest,
            host="127.0.0.1",
            secret="secret",
            bucket_id_salt="bucket_id_salt",
            storage_backend="storage_backend",
            cache_backend="cache_backend",
            permission_backend="permission_backend",
            storage_url="storage_url",
            cache_url="cache_url",
            permission_url="permission_url",
            kinto_version="kinto_version",
            config_file_timestamp="config_file_timestamp",
        )

        self.assertTrue(os.path.exists(dest))

    @mock.patch("kinto.config.render_template")
    def test_hmac_secret_is_text(self, mocked_render_template):
        config.init("kinto.ini", backend="postgresql", cache_backend="postgresql")
        args, kwargs = list(mocked_render_template.call_args)
        self.assertEqual(type(kwargs["secret"]), str)

    @mock.patch("kinto.config.render_template")
    def test_bucket_id_salt_is_text(self, mocked_render_template):
        config.init("kinto.ini", backend="postgresql", cache_backend="postgresql")
        args, kwargs = list(mocked_render_template.call_args)
        self.assertEqual(type(kwargs["bucket_id_salt"]), str)

    @mock.patch("kinto.config.render_template")
    def test_init_postgresql_values(self, mocked_render_template):
        self.maxDiff = None
        config.init("kinto.ini", backend="postgresql", cache_backend="postgresql")

        args, kwargs = list(mocked_render_template.call_args)
        self.assertEqual(args, ("kinto.tpl", "kinto.ini"))

        postgresql_url = "postgresql://postgres:postgres@localhost/postgres"
        self.assertDictEqual(
            kwargs,
            {
                "host": "127.0.0.1",
                "secret": kwargs["secret"],
                "bucket_id_salt": kwargs["bucket_id_salt"],
                "storage_backend": "kinto.core.storage.postgresql",
                "cache_backend": "kinto.core.cache.postgresql",
                "permission_backend": "kinto.core.permission.postgresql",
                "storage_url": postgresql_url,
                "cache_url": postgresql_url,
                "permission_url": postgresql_url,
                "kinto_version": __version__,
                "config_file_timestamp": mock.ANY,
            },
        )

        self._assertTimestampStringsAlmostEqual(
            strftime("%a, %d %b %Y %H:%M:%S %z"),  # expected
            kwargs["config_file_timestamp"],  # actual
        )

    @mock.patch("kinto.config.render_template")
    def test_init_postgresql_memcached_values(self, mocked_render_template):
        config.init("kinto.ini", backend="postgresql", cache_backend="memcached")

        args, kwargs = list(mocked_render_template.call_args)
        self.assertEqual(args, ("kinto.tpl", "kinto.ini"))

        postgresql_url = "postgresql://postgres:postgres@localhost/postgres"
        cache_url = "127.0.0.1:11211 127.0.0.2:11211"
        self.assertDictEqual(
            kwargs,
            {
                "host": "127.0.0.1",
                "secret": kwargs["secret"],
                "bucket_id_salt": kwargs["bucket_id_salt"],
                "storage_backend": "kinto.core.storage.postgresql",
                "cache_backend": "kinto.core.cache.memcached",
                "permission_backend": "kinto.core.permission.postgresql",
                "storage_url": postgresql_url,
                "cache_url": cache_url,
                "permission_url": postgresql_url,
                "kinto_version": __version__,
                "config_file_timestamp": mock.ANY,
            },
        )

        self._assertTimestampStringsAlmostEqual(
            strftime("%a, %d %b %Y %H:%M:%S %z"),  # expected
            kwargs["config_file_timestamp"],  # actual
        )

    @mock.patch("kinto.config.render_template")
    def test_init_memory_values(self, mocked_render_template):
        config.init("kinto.ini", backend="memory", cache_backend="memory")

        args, kwargs = list(mocked_render_template.call_args)
        self.assertEqual(args, ("kinto.tpl", "kinto.ini"))

        self.assertDictEqual(
            kwargs,
            {
                "host": "127.0.0.1",
                "secret": kwargs["secret"],
                "bucket_id_salt": kwargs["bucket_id_salt"],
                "storage_backend": "kinto.core.storage.memory",
                "cache_backend": "kinto.core.cache.memory",
                "permission_backend": "kinto.core.permission.memory",
                "storage_url": "",
                "cache_url": "",
                "permission_url": "",
                "kinto_version": __version__,
                "config_file_timestamp": mock.ANY,
            },
        )

        self._assertTimestampStringsAlmostEqual(
            strftime("%a, %d %b %Y %H:%M:%S %z"),  # expected
            kwargs["config_file_timestamp"],  # actual
        )

    def test_render_template_works_with_file_in_cwd(self):
        temp_path = tempfile.mkdtemp()
        os.chdir(temp_path)
        config.render_template(
            "kinto.tpl",
            "kinto.ini",
            **{
                "host": "127.0.0.1",
                "secret": "abcd-ceci-est-un-secret",
                "bucket_id_salt": "backet-id-salt-random",
                "storage_backend": "kinto.core.storage.memory",
                "cache_backend": "kinto.core.cache.memory",
                "permission_backend": "kinto.core.permission.memory",
                "storage_url": "",
                "cache_url": "",
                "permission_url": "",
                "kinto_version": "",
                "config_file_timestamp": "",
            },
        )
        self.assertTrue(os.path.exists(os.path.join(temp_path, "kinto.ini")))



from unittest import mock

from pyramid.authorization import Authenticated
from pyramid.interfaces import IAuthorizationPolicy
from zope.interface import implementer

from kinto.core import testing
from kinto.core.storage.exceptions import BackendError
from kinto.core.utils import sqlalchemy

from .testapp import main as testapp


# This is the principal a connected user should have (in the tests).
USER_PRINCIPAL = "basicauth:8a931a10fc88ab2f6d1cc02a07d3a81b5d4768f6f13e85c5" "d8d4180419acb1b4"


class BaseWebTest(testing.BaseWebTest):
    """Base Web Test to test your cornice service.

    It setups the database before each test and delete it after.
    """

    entry_point = testapp
    principal = USER_PRINCIPAL

    authorization_policy = "tests.core.support.AllowAuthorizationPolicy"
    plural_url = "/mushrooms"

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.headers.update(testing.get_user_headers("mat"))

    @classmethod
    def get_app_settings(cls, extras=None):
        if extras is None:
            extras = {}
        extras.setdefault("settings_prefix", "myapp")
        extras.setdefault("project_name", "myapp")
        extras.setdefault("project_version", "0.0.1")
        extras.setdefault("http_api_version", "0.1")
        extras.setdefault("project_docs", "https://kinto.readthedocs.io/")
        extras.setdefault("multiauth.policies", "basicauth")
        extras.setdefault("multiauth.authorization_policy", cls.authorization_policy)
        return super().get_app_settings(extras)

    def get_item_url(self, id=None):
        """Return the URL of the item using self.item_url."""
        if id is None:
            id = self.obj["id"]
        return "{}/{}".format(self.plural_url, id)


@implementer(IAuthorizationPolicy)
class AllowAuthorizationPolicy:
    def permits(self, context, principals, permission):
        return Authenticated in principals

    def principals_allowed_by_permission(self, context, permission):
        raise NotImplementedError()  # PRAGMA NOCOVER


class PostgreSQLTest(BaseWebTest):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        if sqlalchemy is not None:
            from .test_cache import PostgreSQLCacheTest
            from .test_permission import PostgreSQLPermissionTest
            from .test_storage import PostgreSQLStorageTest

            settings.update(**PostgreSQLStorageTest.settings)
            settings.update(**PostgreSQLCacheTest.settings)
            settings.update(**PostgreSQLPermissionTest.settings)
            settings.pop("storage_poolclass", None)
            settings.pop("cache_poolclass", None)
            settings.pop("permission_poolclass", None)
        return settings

    def run_failing_batch(self):
        patch = mock.patch.object(self.storage, "delete_all", side_effect=BackendError("boom"))
        self.addCleanup(patch.stop)
        patch.start()
        request_create = {
            "method": "POST",
            "path": "/mushrooms",
            "body": {"data": {"name": "Amanite"}},
        }
        request_delete = {"method": "DELETE", "path": "/mushrooms"}
        body = {"requests": [request_create, request_create, request_delete]}
        self.app.post_json("/batch", body, headers=self.headers, status=503)

    def run_failing_post(self):
        patch = mock.patch.object(
            self.permission, "add_principal_to_ace", side_effect=BackendError("boom")
        )
        self.addCleanup(patch.stop)
        patch.start()
        self.app.post_json(
            "/psilos", {"data": {"name": "Amanite"}}, headers=self.headers, status=503
        )



import requests
from pyramid import authentication as base_auth
from pyramid.interfaces import IAuthenticationPolicy
from pyramid.settings import aslist
from zope.interface import implementer

from kinto.core import logger
from kinto.core import utils as core_utils
from kinto.core.openapi import OpenAPI

from .utils import fetch_openid_config


@implementer(IAuthenticationPolicy)
class OpenIDConnectPolicy(base_auth.CallbackAuthenticationPolicy):
    def __init__(self, issuer, client_id, realm="Realm", **kwargs):
        self.realm = realm
        self.issuer = issuer
        self.client_id = client_id
        self.client_secret = kwargs.get("client_secret", "")
        self.header_type = kwargs.get("header_type", "Bearer")
        self.userid_field = kwargs.get("userid_field", "sub")
        self.verification_ttl = int(kwargs.get("verification_ttl_seconds", 86400))

        # Fetch OpenID config (at instantiation, ie. startup)
        self.oid_config = fetch_openid_config(issuer)

        self._jwt_keys = None

    def unauthenticated_userid(self, request):
        """Return the userid or ``None`` if token could not be verified."""
        settings = request.registry.settings
        hmac_secret = settings["userid_hmac_secret"]

        authorization = request.headers.get("Authorization", "")
        try:
            authmeth, access_token = authorization.split(" ", 1)
        except ValueError:
            return None

        if authmeth.lower() != self.header_type.lower():
            return None

        # XXX JWT Access token
        # https://auth0.com/docs/tokens/access-token#access-token-format

        # Check cache if these tokens were already verified.
        hmac_tokens = core_utils.hmac_digest(hmac_secret, access_token)
        cache_key = f"openid:verify:{hmac_tokens}"
        payload = request.registry.cache.get(cache_key)
        if payload is None:
            # This can take some time.
            payload = self._verify_token(access_token)
            if payload is None:
                return None
        # Save for next time / refresh ttl.
        request.registry.cache.set(cache_key, payload, ttl=self.verification_ttl)
        request.bound_data["user_profile"] = payload
        # Extract meaningful field from userinfo (eg. email or sub)
        return payload.get(self.userid_field)

    def forget(self, request):
        """A no-op. Credentials are sent on every request.
        Return WWW-Authenticate Realm header for Bearer token.
        """
        return [("WWW-Authenticate", '%s realm="%s"' % (self.header_type, self.realm))]

    def _verify_token(self, access_token):
        uri = self.oid_config["userinfo_endpoint"]
        # Opaque access token string. Fetch user info from profile.
        try:
            resp = requests.get(uri, headers={"Authorization": "Bearer " + access_token})
            resp.raise_for_status()
            userprofile = resp.json()
            return userprofile

        except (requests.exceptions.HTTPError, ValueError, KeyError) as e:
            logger.debug("Unable to fetch user profile from %s (%s)" % (uri, e))
            return None


def get_user_profile(request):
    return request.bound_data.get("user_profile", {})


def includeme(config):
    # Activate end-points.
    config.scan("kinto.plugins.openid.views")

    settings = config.get_settings()

    openid_policies = []
    for policy in aslist(settings["multiauth.policies"]):
        v = settings.get("multiauth.policy.%s.use" % policy, "")
        if v.endswith("OpenIDConnectPolicy"):
            openid_policies.append(policy)

    if len(openid_policies) == 0:
        # Do not add the capability if no policy is configured.
        return

    providers_infos = []
    for name in openid_policies:
        issuer = settings["multiauth.policy.%s.issuer" % name]
        openid_config = fetch_openid_config(issuer)

        client_id = settings["multiauth.policy.%s.client_id" % name]
        header_type = settings.get("multiauth.policy.%s.header_type", "Bearer")

        providers_infos.append(
            {
                "name": name,
                "issuer": openid_config["issuer"],
                "auth_path": "/openid/%s/login" % name,
                "client_id": client_id,
                "header_type": header_type,
                "userinfo_endpoint": openid_config["userinfo_endpoint"],
            }
        )

        OpenAPI.expose_authentication_method(
            name, {"type": "oauth2", "authorizationUrl": openid_config["authorization_endpoint"]}
        )

    config.add_api_capability(
        "openid",
        description="OpenID connect support.",
        url="http://kinto.readthedocs.io/en/stable/api/1.x/authentication.html",
        providers=providers_infos,
    )
    config.add_request_method(get_user_profile, name="get_user_profile")



import unittest
from unittest import mock

from kinto.core.testing import DummyRequest
from kinto.plugins.openid import OpenIDConnectPolicy
from kinto.plugins.openid.utils import fetch_openid_config

from .. import support


def get_openid_configuration(url):
    base_url = url.replace("/.well-known/openid-configuration", "")
    m = mock.Mock()
    m.json.return_value = {
        "issuer": "{base_url} issuer".format(base_url=base_url),
        "authorization_endpoint": "{base_url}/authorize".format(base_url=base_url),
        "userinfo_endpoint": "{base_url}/oauth/user".format(base_url=base_url),
        "token_endpoint": "{base_url}/oauth/token".format(base_url=base_url),
    }
    return m


class OpenIDWebTest(support.BaseWebTest, unittest.TestCase):
    @classmethod
    def make_app(cls, *args, **kwargs):
        with mock.patch("kinto.plugins.openid.requests.get") as get:
            get.side_effect = get_openid_configuration
            return super(OpenIDWebTest, cls).make_app(*args, **kwargs)

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        openid_policy = "kinto.plugins.openid.OpenIDConnectPolicy"
        settings["includes"] = "kinto.plugins.openid"
        settings["multiauth.policies"] = "auth0 google"
        settings["multiauth.policy.auth0.use"] = openid_policy
        settings["multiauth.policy.auth0.issuer"] = "https://auth.mozilla.auth0.com"
        settings["multiauth.policy.auth0.client_id"] = "abc"
        settings["multiauth.policy.auth0.client_secret"] = "xyz"

        settings["multiauth.policy.google.use"] = openid_policy
        settings["multiauth.policy.google.issuer"] = "https://accounts.google.com"
        settings["multiauth.policy.google.client_id"] = "123"
        settings["multiauth.policy.google.client_secret"] = "789"
        settings["multiauth.policy.google.userid_field"] = "email"
        return settings

    def test_openid_multiple_providers(self):
        resp = self.app.get("/")
        capabilities = resp.json["capabilities"]
        providers = capabilities["openid"]["providers"]
        assert len(providers) == 2


class OpenIDWithoutPolicyTest(support.BaseWebTest, unittest.TestCase):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["includes"] = "kinto.plugins.openid"
        return settings

    def test_openid_capability_is_not_added(self):
        resp = self.app.get("/")
        capabilities = resp.json["capabilities"]
        assert "openid" not in capabilities


class OpenIDOnePolicyTest(support.BaseWebTest, unittest.TestCase):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        openid_policy = "kinto.plugins.openid.OpenIDConnectPolicy"
        settings["includes"] = "kinto.plugins.openid"
        settings["multiauth.policies"] = "google"
        settings["multiauth.policy.auth0.use"] = openid_policy
        settings["multiauth.policy.auth0.issuer"] = "https://auth.mozilla.auth0.com"
        settings["multiauth.policy.auth0.client_id"] = "abc"
        settings["multiauth.policy.auth0.client_secret"] = "xyz"

        settings["multiauth.policy.google.use"] = openid_policy
        settings["multiauth.policy.google.issuer"] = "https://accounts.google.com"
        settings["multiauth.policy.google.client_id"] = "123"
        settings["multiauth.policy.google.client_secret"] = "789"
        settings["multiauth.policy.google.userid_field"] = "email"
        return settings

    def test_openid_one_provider(self):
        resp = self.app.get("/")
        capabilities = resp.json["capabilities"]
        providers = capabilities["openid"]["providers"]
        assert len(providers) == 1

    def test_profile_is_exposed(self):
        key = "openid:verify:444c6694937007bbf494f155f6cb12139db4c4c6a926742f3fe0bb4b5d191aa3"
        profile = {"sub": "abcd", "email": "foobar@tld.com"}
        self.app.app.registry.cache.set(key, profile, ttl=30)
        with mock.patch("kinto.plugins.openid.utils.requests.get") as m:
            m.return_value.json.return_value = {
                "userinfo_endpoint": "http://uinfo",
                "jwks_uri": "https://jwks",
            }
            fetch_openid_config("https://fxa")

        resp = self.app.get("/", headers={"Authorization": "Bearer avrbnnbrbr"})
        assert "profile" in resp.json["user"]
        assert resp.json["user"]["profile"] == {"sub": "abcd", "email": "foobar@tld.com"}


class HelloViewTest(OpenIDWebTest):
    def test_openid_capability_if_enabled(self):
        resp = self.app.get("/")
        capabilities = resp.json["capabilities"]
        assert "openid" in capabilities
        assert len(capabilities["openid"]["providers"]) == 2
        assert "userinfo_endpoint" in capabilities["openid"]["providers"][0]
        assert "auth_path" in capabilities["openid"]["providers"][0]

    def test_openid_in_openapi(self):
        resp = self.app.get("/__api__")
        assert "auth0" in resp.json["securityDefinitions"]
        auth = resp.json["securityDefinitions"]["auth0"]["authorizationUrl"]
        assert auth == "https://auth.mozilla.auth0.com/authorize"


class PolicyTest(unittest.TestCase):
    def setUp(self):
        mocked = mock.patch("kinto.plugins.openid.requests.get")
        self.mocked_get = mocked.start()
        self.addCleanup(mocked.stop)

        self.policy = OpenIDConnectPolicy(issuer="https://idp", client_id="abc")

        self.request = DummyRequest()
        self.request.registry.cache.get.return_value = None

        mocked = mock.patch.object(self.policy, "_verify_token")
        self.verify = mocked.start()
        self.addCleanup(mocked.stop)
        self.verify.return_value = {"sub": "userid"}

    def test_returns_none_if_no_authorization(self):
        assert self.policy.unauthenticated_userid(self.request) is None

    def test_returns_header_type_in_forget(self):
        h = self.policy.forget(self.request)
        assert "Bearer " in h[0][1]

    def test_header_type_can_be_configured(self):
        self.policy.header_type = "bearer+oidc"
        h = self.policy.forget(self.request)
        assert "bearer+oidc " in h[0][1]

    def test_returns_none_if_no_authorization_prefix(self):
        self.request.headers["Authorization"] = "avrbnnbrbr"
        assert self.policy.unauthenticated_userid(self.request) is None

    def test_returns_none_if_bad_prefix(self):
        self.request.headers["Authorization"] = "Basic avrbnnbrbr"
        assert self.policy.unauthenticated_userid(self.request) is None

    def test_can_specify_only_opaque_access_token(self):
        self.request.headers["Authorization"] = "Bearer xyz"
        assert self.policy.unauthenticated_userid(self.request) == "userid"
        self.verify.assert_called_with("xyz")

    def test_returns_none_if_no_cache_and_invalid_access_token(self):
        self.request.headers["Authorization"] = "Bearer xyz"
        self.request.registry.cache.get.return_value = None
        self.verify.return_value = None
        assert self.policy.unauthenticated_userid(self.request) is None
        assert not self.request.registry.cache.set.called

    def test_payload_is_read_from_cache(self):
        self.request.headers["Authorization"] = "Bearer xyz"
        self.request.registry.cache.get.return_value = {"sub": "me"}
        assert self.policy.unauthenticated_userid(self.request) == "me"

    def test_payload_is_stored_in_cache(self):
        self.request.headers["Authorization"] = "Bearer xyz"
        assert self.policy.unauthenticated_userid(self.request) == "userid"
        assert self.request.registry.cache.set.called

    def test_payload_is_read_from_cache_but_differently_by_access_token(self):
        # State to keep track of cache keys queried.
        cache_keys_used = []

        def mocked_cache_get(cache_key):
            # This makes sure the same cache key is not used twice
            assert cache_key not in cache_keys_used
            cache_keys_used.append(cache_key)
            if len(cache_keys_used) == 1:
                return {"sub": "me"}
            elif len(cache_keys_used) == 2:
                return {"sub": "you"}

        self.request.registry.cache.get.side_effect = mocked_cache_get

        self.request.headers["Authorization"] = "Bearer xyz"
        assert self.policy.unauthenticated_userid(self.request) == "me"

        # Change the Authorization header the second time
        self.request.headers["Authorization"] = "Bearer abc"
        assert self.policy.unauthenticated_userid(self.request) == "you"


class VerifyTokenTest(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        # Populate OpenID config cache.
        with mock.patch("kinto.plugins.openid.utils.requests.get") as m:
            m.return_value.json.return_value = {
                "userinfo_endpoint": "http://uinfo",
                "jwks_uri": "https://jwks",
            }
            fetch_openid_config("https://fxa")

    def setUp(self):
        mocked = mock.patch("kinto.plugins.openid.requests.get")
        self.mocked_get = mocked.start()
        self.addCleanup(mocked.stop)

        self.policy = OpenIDConnectPolicy(issuer="https://fxa", client_id="abc")

    def test_fetches_userinfo_if_id_token_is_none(self):
        self.mocked_get.return_value.json.side_effect = [{"sub": "me"}]
        payload = self.policy._verify_token(access_token="abc")
        assert payload["sub"] == "me"

    def test_returns_none_if_fetching_userinfo_fails(self):
        self.mocked_get.return_value.raise_for_status.side_effect = ValueError
        payload = self.policy._verify_token(access_token="abc")
        assert payload is None


class LoginViewTest(OpenIDWebTest):
    def test_returns_400_if_parameters_are_missing_or_bad(self):
        self.app.get("/openid/auth0/login", status=400)
        self.app.get("/openid/auth0/login", params={"callback": "http://no-scope"}, status=400)
        self.app.get(
            "/openid/auth0/login", params={"callback": "bad", "scope": "openid"}, status=400
        )

    def test_returns_400_if_provider_is_unknown(self):
        self.app.get("/openid/fxa/login", status=400)

    def test_returns_400_if_email_is_not_in_scope_when_userid_field_is_email(self):
        scope = "openid"
        cb = "http://ui.kinto.example.com"
        self.app.get("/openid/auth0/login", params={"callback": cb, "scope": scope}, status=307)
        # See config above (email is userid field)
        self.app.get("/openid/google/login", params={"callback": cb, "scope": scope}, status=400)

    def test_returns_400_if_prompt_is_not_recognized(self):
        scope = "openid"
        cb = "http://ui"
        self.app.get(
            "/openid/auth0/login",
            params={"callback": cb, "scope": scope, "prompt": "junk"},
            status=400,
        )

    def test_redirects_to_the_identity_provider(self):
        params = {"callback": "http://ui.kinto.example.com", "scope": "openid"}
        resp = self.app.get("/openid/auth0/login", params=params, status=307)
        location = resp.headers["Location"]
        assert "auth0.com/authorize?" in location
        assert "%2Fv1%2Fopenid%2Fauth0%2Ftoken" in location
        assert "scope=openid" in location
        assert "client_id=abc" in location

    def test_redirects_to_the_identity_provider_with_prompt_none(self):
        params = {"callback": "http://ui.kinto.example.com", "scope": "openid", "prompt": "none"}
        resp = self.app.get("/openid/auth0/login", params=params, status=307)
        location = resp.headers["Location"]
        assert "auth0.com/authorize?" in location
        assert "%2Fv1%2Fopenid%2Fauth0%2Ftoken" in location
        assert "scope=openid" in location
        assert "client_id=abc" in location
        assert "prompt=none" in location

    def test_callback_is_stored_in_cache(self):
        params = {"callback": "http://ui.kinto.example.com", "scope": "openid"}
        with mock.patch("kinto.plugins.openid.views.random_bytes_hex") as m:
            m.return_value = "key"
            self.app.get("/openid/auth0/login", params=params, status=307)

        cached = self.app.app.registry.cache.get("openid:state:key")
        assert cached == "http://ui.kinto.example.com"


class TokenViewTest(OpenIDWebTest):
    def test_returns_400_if_parameters_are_missing_or_bad(self):
        self.app.get("/openid/auth0/token", status=400)
        self.app.get("/openid/auth0/token", params={"code": "abc"}, status=400)
        self.app.get("/openid/auth0/token", params={"state": "abc"}, status=400)

    def test_returns_400_if_provider_is_unknown(self):
        self.app.get("/openid/fxa/token", status=400)

    def test_returns_400_if_state_is_invalid(self):
        self.app.get("/openid/auth0/token", params={"code": "abc", "state": "abc"}, status=400)

    def test_code_is_traded_using_client_secret(self):
        self.app.app.registry.cache.set("openid:state:key", "http://ui", ttl=100)
        with mock.patch("kinto.plugins.openid.views.requests.post") as m:
            m.return_value.text = '{"access_token": "token"}'
            self.app.get("/openid/auth0/token", params={"code": "abc", "state": "key"})
            m.assert_called_with(
                "https://auth.mozilla.auth0.com/oauth/token",
                data={
                    "code": "abc",
                    "client_id": "abc",
                    "client_secret": "xyz",
                    "redirect_uri": "http://localhost/v1/openid/auth0/token",
                    "grant_type": "authorization_code",
                },
            )

    def test_state_cannot_be_reused(self):
        self.app.app.registry.cache.set("openid:state:key", "http://ui", ttl=100)
        with mock.patch("kinto.plugins.openid.views.requests.post") as m:
            m.return_value.text = '{"access_token": "token"}'
            self.app.get("/openid/auth0/token", params={"code": "abc", "state": "key"})
            self.app.get("/openid/auth0/token", params={"code": "abc", "state": "key"}, status=400)

    def test_redirects_to_callback_using_authorization_response(self):
        self.app.app.registry.cache.set("openid:state:key", "http://ui/#token=", ttl=100)
        with mock.patch("kinto.plugins.openid.views.requests.post") as m:
            m.return_value.text = '{"access_token": "token"}'
            resp = self.app.get(
                "/openid/auth0/token", params={"code": "abc", "state": "key"}, status=307
            )
        location = resp.headers["Location"]
        assert location == "http://ui/#token=eyJhY2Nlc3NfdG9rZW4iOiAidG9rZW4ifQ%3D%3D"
